from dataclasses import dataclass
from abc import ABC
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty


@dataclass(kw_only=True)
class Rock(ABC):
    """The abstract base class for a collection of rock property inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of rock property inputs, as a dictionary.
    """

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """A Collection of rock property inputs, as a dictionary."""
        raise NotImplementedError("Implement this in the derived class")

    def to_string(self) -> str:
        """Writes dynamic property data to string."""
        raise NotImplementedError('Implement this in the derived class.')



"""Nexus grid file class for loading in a structured grid file and extracting the grid properties."""
from __future__ import annotations

import copy

import pandas as pd
from dataclasses import dataclass, field
from typing import Optional, TYPE_CHECKING, Any, Final
import warnings

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.DataModelBaseClasses.Grid import Grid
from ResSimpy.DataModelBaseClasses.GridArrayDefinition import GridArrayDefinition
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusFtrans import NexusFtrans
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusGridArrayFunction import NexusGridArrayFunction
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusLGRs import NexusLGRs
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusMultir import NexusMultir
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusOver import NexusOver
from ResSimpy.Nexus.DataModels.StructuredGrid.NexusTOver import NexusTOver
from ResSimpy.Nexus.NexusKeywords.nexus_keywords import VALID_NEXUS_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_FORMAT_KEYWORDS
from ResSimpy.Nexus.structured_grid_operations import StructuredGridOperations
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.Nexus.array_function_operations as afo
from ResSimpy.FileOperations import file_operations as fo
from resqpy.olio.read_nexus_fault import load_nexus_fault_mult_table_from_list

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass
class PropertyToLoad:
    token: str
    modifiers: list[str]
    property: GridArrayDefinition | dict[str, GridArrayDefinition]


@dataclass(kw_only=True)
class NexusGrid(Grid):
    __array_functions_list: Optional[list[list[str]]] = None
    __array_functions_df: Optional[pd.DataFrame] = None
    __array_functions_loaded: bool = False
    __grid_file_contents: Optional[list[str]] = None
    __grid_file_nested: Optional[list[str]] = None
    __faults_df: Optional[pd.DataFrame] = None
    __grid_faults_loaded: bool = False
    __grid_nexus_file: Optional[NexusFile] = None
    __corp: GridArrayDefinition
    __dx: GridArrayDefinition
    __dy: GridArrayDefinition
    __dz: GridArrayDefinition
    __depth: GridArrayDefinition
    __mdepth: GridArrayDefinition
    __dznet: GridArrayDefinition
    __compr: GridArrayDefinition
    __icoars: GridArrayDefinition
    __ialphaf: GridArrayDefinition
    __ipolymer: GridArrayDefinition
    __iadsorption: GridArrayDefinition
    __itracer: GridArrayDefinition
    __igrid: GridArrayDefinition
    __isector: GridArrayDefinition
    __swl: GridArrayDefinition
    __swr: GridArrayDefinition
    __swu: GridArrayDefinition
    __sgl: GridArrayDefinition
    __sgr: GridArrayDefinition
    __sgu: GridArrayDefinition
    __swro: GridArrayDefinition
    __swro_ls: GridArrayDefinition
    __sgro: GridArrayDefinition
    __sgrw: GridArrayDefinition
    __krw_swro: GridArrayDefinition
    __krws_ls: GridArrayDefinition
    __krw_swu: GridArrayDefinition
    __krg_sgro: GridArrayDefinition
    __krg_sgu: GridArrayDefinition
    __krg_sgrw: GridArrayDefinition
    __kro_swl: GridArrayDefinition
    __kro_swr: GridArrayDefinition
    __kro_sgl: GridArrayDefinition
    __kro_sgr: GridArrayDefinition
    __krw_sgl: GridArrayDefinition
    __krw_sgr: GridArrayDefinition
    __sgtr: GridArrayDefinition
    __sotr: GridArrayDefinition
    __swlpc: GridArrayDefinition
    __sglpc: GridArrayDefinition
    __pcw_swl: GridArrayDefinition
    __pcg_sgu: GridArrayDefinition
    __chloride: GridArrayDefinition
    __calcium: GridArrayDefinition
    __salinity: GridArrayDefinition
    __api: GridArrayDefinition
    __tmx: GridArrayDefinition
    __tmy: GridArrayDefinition
    __tmz: GridArrayDefinition
    __multbv: GridArrayDefinition
    __pv: GridArrayDefinition
    __iequil: GridArrayDefinition
    __ipvt: GridArrayDefinition
    __iwater: GridArrayDefinition
    __irelpm: GridArrayDefinition
    __irock: GridArrayDefinition
    __itran: GridArrayDefinition
    _iregion: dict[str, GridArrayDefinition]
    __pvmult: GridArrayDefinition
    __livecell: GridArrayDefinition
    __worka1: GridArrayDefinition
    __worka2: GridArrayDefinition
    __worka3: GridArrayDefinition
    __worka4: GridArrayDefinition
    __worka5: GridArrayDefinition
    __worka6: GridArrayDefinition
    __worka7: GridArrayDefinition
    __worka8: GridArrayDefinition
    __worka9: GridArrayDefinition
    __kxeff: GridArrayDefinition
    __kyeff: GridArrayDefinition
    __kzeff: GridArrayDefinition
    __grid_multir_loaded: bool = False
    __multir: Optional[list[NexusMultir]] = None
    __lgrs: NexusLGRs
    __overs: list[NexusOver] = field(default_factory=list)
    __tovers: list[NexusTOver] = field(default_factory=list)
    __ftrans: list[NexusFtrans] = field(default_factory=list)
    __model_unit_system: UnitSystem

    def __init__(self, model_unit_system: UnitSystem, grid_nexus_file: Optional[NexusFile] = None,
                 assume_loaded: bool = False,
                 ) -> None:
        """Initialises the NexusGrid class.

        Args:
            model_unit_system (UnitSystem): the default unit system to use for the model.
            grid_nexus_file (Optional[NexusFile]): the NexusFile representation of a structured grid file for \
                reading and interpreting the grid properties from.
            assume_loaded (bool): Create the object assuming the grid has already been loaded into memory.
        """
        super().__init__(assume_loaded=assume_loaded)
        self.__array_functions_list: Optional[list[list[str]]] = None
        self.__array_functions_df: Optional[pd.DataFrame] = None
        self.__array_functions_loaded: bool = False
        self.__grid_file_contents: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.get_flat_list_str_file_including_includes
        self.__grid_file_nested: Optional[list[str]] = None if grid_nexus_file is None else \
            grid_nexus_file.file_content_as_list
        self.__faults_df: Optional[pd.DataFrame] = None
        self.__grid_nexus_file: Optional[NexusFile] = grid_nexus_file
        self.__grid_array_functions: Optional[list[NexusGridArrayFunction]] = None
        self.__corp: GridArrayDefinition = GridArrayDefinition()
        self.__iequil: GridArrayDefinition = GridArrayDefinition()
        self.__ipvt: GridArrayDefinition = GridArrayDefinition()
        self.__iwater: GridArrayDefinition = GridArrayDefinition()
        self.__irelpm: GridArrayDefinition = GridArrayDefinition()
        self.__irock: GridArrayDefinition = GridArrayDefinition()
        self.__itran: GridArrayDefinition = GridArrayDefinition()
        self._iregion: dict[str, GridArrayDefinition] = {}
        self.__pvmult: GridArrayDefinition = GridArrayDefinition()
        self.__livecell: GridArrayDefinition = GridArrayDefinition()
        self.__worka1: GridArrayDefinition = GridArrayDefinition()
        self.__worka2: GridArrayDefinition = GridArrayDefinition()
        self.__worka3: GridArrayDefinition = GridArrayDefinition()
        self.__worka4: GridArrayDefinition = GridArrayDefinition()
        self.__worka5: GridArrayDefinition = GridArrayDefinition()
        self.__worka6: GridArrayDefinition = GridArrayDefinition()
        self.__worka7: GridArrayDefinition = GridArrayDefinition()
        self.__worka8: GridArrayDefinition = GridArrayDefinition()
        self.__worka9: GridArrayDefinition = GridArrayDefinition()
        self.__dx: GridArrayDefinition = GridArrayDefinition()
        self.__dy: GridArrayDefinition = GridArrayDefinition()
        self.__dz: GridArrayDefinition = GridArrayDefinition()
        self.__depth: GridArrayDefinition = GridArrayDefinition()
        self.__mdepth: GridArrayDefinition = GridArrayDefinition()
        self.__dznet: GridArrayDefinition = GridArrayDefinition()
        self.__compr: GridArrayDefinition = GridArrayDefinition()
        self.__icoars: GridArrayDefinition = GridArrayDefinition()
        self.__ialphaf: GridArrayDefinition = GridArrayDefinition()
        self.__ipolymer: GridArrayDefinition = GridArrayDefinition()
        self.__iadsorption: GridArrayDefinition = GridArrayDefinition()
        self.__itracer: GridArrayDefinition = GridArrayDefinition()
        self.__igrid: GridArrayDefinition = GridArrayDefinition()
        self.__isector: GridArrayDefinition = GridArrayDefinition()
        self.__swl: GridArrayDefinition = GridArrayDefinition()
        self.__swr: GridArrayDefinition = GridArrayDefinition()
        self.__swu: GridArrayDefinition = GridArrayDefinition()
        self.__sgl: GridArrayDefinition = GridArrayDefinition()
        self.__sgr: GridArrayDefinition = GridArrayDefinition()
        self.__sgu: GridArrayDefinition = GridArrayDefinition()
        self.__swro: GridArrayDefinition = GridArrayDefinition()
        self.__swro_ls: GridArrayDefinition = GridArrayDefinition()
        self.__sgro: GridArrayDefinition = GridArrayDefinition()
        self.__sgrw: GridArrayDefinition = GridArrayDefinition()
        self.__krw_swro: GridArrayDefinition = GridArrayDefinition()
        self.__krws_ls: GridArrayDefinition = GridArrayDefinition()
        self.__krw_swu: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgro: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgu: GridArrayDefinition = GridArrayDefinition()
        self.__krg_sgrw: GridArrayDefinition = GridArrayDefinition()
        self.__kro_swl: GridArrayDefinition = GridArrayDefinition()
        self.__kro_swr: GridArrayDefinition = GridArrayDefinition()
        self.__kro_sgl: GridArrayDefinition = GridArrayDefinition()
        self.__kro_sgr: GridArrayDefinition = GridArrayDefinition()
        self.__krw_sgl: GridArrayDefinition = GridArrayDefinition()
        self.__krw_sgr: GridArrayDefinition = GridArrayDefinition()
        self.__sgtr: GridArrayDefinition = GridArrayDefinition()
        self.__sotr: GridArrayDefinition = GridArrayDefinition()
        self.__swlpc: GridArrayDefinition = GridArrayDefinition()
        self.__sglpc: GridArrayDefinition = GridArrayDefinition()
        self.__pcw_swl: GridArrayDefinition = GridArrayDefinition()
        self.__pcg_sgu: GridArrayDefinition = GridArrayDefinition()
        self.__chloride: GridArrayDefinition = GridArrayDefinition()
        self.__calcium: GridArrayDefinition = GridArrayDefinition()
        self.__salinity: GridArrayDefinition = GridArrayDefinition()
        self.__api: GridArrayDefinition = GridArrayDefinition()
        self.__tmx: GridArrayDefinition = GridArrayDefinition()
        self.__tmy: GridArrayDefinition = GridArrayDefinition()
        self.__tmz: GridArrayDefinition = GridArrayDefinition()
        self.__multbv: GridArrayDefinition = GridArrayDefinition()
        self.__pv: GridArrayDefinition = GridArrayDefinition()
        self.__kxeff: GridArrayDefinition = GridArrayDefinition()
        self.__kyeff: GridArrayDefinition = GridArrayDefinition()
        self.__kzeff: GridArrayDefinition = GridArrayDefinition()

        self.__lgrs: NexusLGRs = NexusLGRs(grid_file_as_list=self.__grid_file_contents, parent_grid=self)
        self.__overs: list[NexusOver] = []
        self.__tovers: list[NexusTOver] = []
        self.__ftrans: list[NexusFtrans] = []
        self.__model_unit_system: UnitSystem = model_unit_system

    def __wrap(self, value: Any) -> Any:
        if isinstance(value, tuple | list | set | frozenset):
            return type(value)([self.__wrap(v) for v in value])
        else:
            return value

    def update_properties_from_dict(self, data: dict[str, int | GridArrayDefinition]) -> None:
        """Allows you to update properties on the class using the provided dict of values.

        Args:
        ----
                data (dict[str, int | GridArrayDefinition]): the dictionary of values to update on the class
        """
        # Use the dict provided to populate the properties in the class
        if data is not None:
            for name, value in data.items():
                private_name = '_' + name
                setattr(self, private_name, self.__wrap(value))

        # Prevent reload from disk
        self._grid_properties_loaded = True

    def to_dict(self) -> dict[str, Optional[int] | GridArrayDefinition]:
        """Returns a dictionary representation of grid array.
        Checks if grid properties are not loaded and loads them.
        """
        self.load_grid_properties_if_not_loaded()

        original_dict = self.__dict__
        new_dict: dict[str, Optional[int] | GridArrayDefinition] = {}
        for key in original_dict.keys():
            new_key = key
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            if new_key[0] == '_':
                new_key = new_key.replace('_', '', 1)
            new_dict[new_key] = original_dict[key]

        return new_dict

    @staticmethod
    def keyword_mapping() -> dict[str, tuple[str, type]]:
        """Returns the keyword mapping for the NexusGrid class."""
        keyword_map = {
            'NETGRS': ('netgrs', GridArrayDefinition),
            'POROSITY': ('porosity', GridArrayDefinition),
            'POR': ('porosity', GridArrayDefinition),
            'SW': ('sw', GridArrayDefinition),
            'SG': ('sg', GridArrayDefinition),
            'PRESSURE': ('pressure', GridArrayDefinition),
            'P': ('pressure', GridArrayDefinition),
            'TEMPERATURE': ('temperature', GridArrayDefinition),
            'TEMP': ('temperature', GridArrayDefinition),
            'KX': ('kx', GridArrayDefinition),
            'KI': ('kx', GridArrayDefinition),
            'PERMX': ('kx', GridArrayDefinition),
            'PERMI': ('kx', GridArrayDefinition),
            'KY': ('ky', GridArrayDefinition),
            'KXEFF': ('kxeff', GridArrayDefinition),
            'KYEFF': ('kyeff', GridArrayDefinition),
            'KZEFF': ('kzeff', GridArrayDefinition),
            'KJ': ('ky', GridArrayDefinition),
            'PERMY': ('ky', GridArrayDefinition),
            'PERMJ': ('ky', GridArrayDefinition),
            'KZ': ('kz', GridArrayDefinition),
            'KK': ('kz', GridArrayDefinition),
            'PERMZ': ('kz', GridArrayDefinition),
            'PERMK': ('kz', GridArrayDefinition),
            'IREGION': ('iregion', dict),
            'PVMULT': ('pvmult', GridArrayDefinition),
            'CORP': ('corp', GridArrayDefinition),
            'IEQUIL': ('iequil', GridArrayDefinition),
            'IPVT': ('ipvt', GridArrayDefinition),
            'IWATER': ('iwater', GridArrayDefinition),
            'IRELPM': ('irelpm', GridArrayDefinition),
            'IROCK': ('irock', GridArrayDefinition),
            'ITRAN': ('itran', GridArrayDefinition),
            'LIVECELL': ('livecell', GridArrayDefinition),
            'WORKA1': ('worka1', GridArrayDefinition),
            'WORKA2': ('worka2', GridArrayDefinition),
            'WORKA3': ('worka3', GridArrayDefinition),
            'WORKA4': ('worka4', GridArrayDefinition),
            'WORKA5': ('worka5', GridArrayDefinition),
            'WORKA6': ('worka6', GridArrayDefinition),
            'WORKA7': ('worka7', GridArrayDefinition),
            'WORKA8': ('worka8', GridArrayDefinition),
            'WORKA9': ('worka9', GridArrayDefinition),
            'DX': ('dx', GridArrayDefinition),
            'DY': ('dy', GridArrayDefinition),
            'DZ': ('dz', GridArrayDefinition),
            'DEPTH': ('depth', GridArrayDefinition),
            'MDEPTH': ('mdepth', GridArrayDefinition),
            'DZNET': ('dznet', GridArrayDefinition),
            'COMPR': ('compr', GridArrayDefinition),
            'ICOARS': ('icoars', GridArrayDefinition),
            'IALPHAF': ('ialphaf', GridArrayDefinition),
            'IPOLYMER': ('ipolymer', GridArrayDefinition),
            'IADSORPTION': ('iadsorption', GridArrayDefinition),
            'ITRACER': ('itracer', GridArrayDefinition),
            'IGRID': ('igrid', GridArrayDefinition),
            'ISECTOR': ('isector', GridArrayDefinition),
            'SWL': ('swl', GridArrayDefinition),
            'SWR': ('swr', GridArrayDefinition),
            'SWU': ('swu', GridArrayDefinition),
            'SGL': ('sgl', GridArrayDefinition),
            'SGR': ('sgr', GridArrayDefinition),
            'SGU': ('sgu', GridArrayDefinition),
            'SWRO': ('swro', GridArrayDefinition),
            'SWRO_LS': ('swro_ls', GridArrayDefinition),
            'SGRO': ('sgro', GridArrayDefinition),
            'SGRW': ('sgrw', GridArrayDefinition),
            'KRW_SWRO': ('krw_swro', GridArrayDefinition),
            'KRWS_LS': ('krws_ls', GridArrayDefinition),
            'KRW_SWU': ('krw_swu', GridArrayDefinition),
            'KRG_SGRO': ('krg_sgro', GridArrayDefinition),
            'KRG_SGU': ('krg_sgu', GridArrayDefinition),
            'KRG_SGRW': ('krg_sgrw', GridArrayDefinition),
            'KRO_SWL': ('kro_swl', GridArrayDefinition),
            'KRO_SWR': ('kro_swr', GridArrayDefinition),
            'KRO_SGL': ('kro_sgl', GridArrayDefinition),
            'KRO_SGR': ('kro_sgr', GridArrayDefinition),
            'KRW_SGL': ('krw_sgl', GridArrayDefinition),
            'KRW_SGR': ('krw_sgr', GridArrayDefinition),
            'SGTR': ('sgtr', GridArrayDefinition),
            'SOTR': ('sotr', GridArrayDefinition),
            'SWLPC': ('swlpc', GridArrayDefinition),
            'SGLPC': ('sglpc', GridArrayDefinition),
            'PCW_SWL': ('pcw_swl', GridArrayDefinition),
            'PCG_SGU': ('pcg_sgu', GridArrayDefinition),
            'CHLORIDE': ('chloride', GridArrayDefinition),
            'CALCIUM': ('calcium', GridArrayDefinition),
            'SALINITY': ('salinity', GridArrayDefinition),
            'API': ('api', GridArrayDefinition),
            'TMX': ('tmx', GridArrayDefinition),
            'TMY': ('tmy', GridArrayDefinition),
            'TMZ': ('tmz', GridArrayDefinition),
            'MULTBV': ('multbv', GridArrayDefinition),
            'PV': ('pv', GridArrayDefinition),

        }
        return keyword_map

    def load_grid_properties_if_not_loaded(self) -> None:
        """Checks if grid properties are not loaded and loads them."""

        def move_next_value(next_line: str) -> tuple[str, str]:
            """Finds the next value and then strips out the value from the line.

            Args:
            ----
                next_line (str): the line to search through for the value

            Raises:
            ------
                ValueError: if no value is found within the line provided

            Returns:
            -------
                tuple[str, str]: the next value found in the line, the line with the value stripped out.
            """
            value = nfo.get_next_value(0, [next_line], next_line)
            if value is None:
                raise ValueError(f"No value found within the provided line: {next_line}")
            next_line = next_line.replace(value, "", 1)
            return value, next_line

        # If we've already loaded the grid properties, don't do so again.
        if self._grid_properties_loaded:
            return

        self.lgrs.load_lgrs()

        if self.__grid_nexus_file is None or self.__grid_file_contents is None or self.__grid_file_nested is None:
            raise ValueError("Grid file not found, cannot load grid properties")

        # Strip file of comments
        file_as_list_with_original_line_numbers = []
        for i, line in enumerate(self.__grid_file_contents):
            cleaned_line = nfo.strip_file_of_comments([line], comment_characters=['!', 'C'])
            if len(cleaned_line) == 0 or cleaned_line[0].strip() == '':
                pass
            else:
                file_as_list_with_original_line_numbers.append((i, cleaned_line[0]))

        file_as_list = [line for _, line in file_as_list_with_original_line_numbers]

        properties_to_load = [
            PropertyToLoad('NETGRS', GRID_ARRAY_FORMAT_KEYWORDS, self._netgrs),
            PropertyToLoad('POROSITY', GRID_ARRAY_FORMAT_KEYWORDS, self._porosity),
            PropertyToLoad('POR', GRID_ARRAY_FORMAT_KEYWORDS, self._porosity),
            PropertyToLoad('SW', GRID_ARRAY_FORMAT_KEYWORDS, self._sw),
            PropertyToLoad('SG', GRID_ARRAY_FORMAT_KEYWORDS, self._sg),
            PropertyToLoad('PRESSURE', GRID_ARRAY_FORMAT_KEYWORDS, self._pressure),
            PropertyToLoad('P', GRID_ARRAY_FORMAT_KEYWORDS, self._pressure),
            PropertyToLoad('TEMPERATURE', GRID_ARRAY_FORMAT_KEYWORDS, self._temperature),
            PropertyToLoad('TEMP', GRID_ARRAY_FORMAT_KEYWORDS, self._temperature),
            PropertyToLoad('KX', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('KI', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('PERMX', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('PERMI', GRID_ARRAY_FORMAT_KEYWORDS, self._kx),
            PropertyToLoad('KY', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('KXEFF', GRID_ARRAY_FORMAT_KEYWORDS, self.__kxeff),
            PropertyToLoad('KYEFF', GRID_ARRAY_FORMAT_KEYWORDS, self.__kyeff),
            PropertyToLoad('KZEFF', GRID_ARRAY_FORMAT_KEYWORDS, self.__kzeff),
            PropertyToLoad('KJ', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('PERMY', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('PERMJ', GRID_ARRAY_FORMAT_KEYWORDS, self._ky),
            PropertyToLoad('KZ', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('KK', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('PERMZ', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('PERMK', GRID_ARRAY_FORMAT_KEYWORDS, self._kz),
            PropertyToLoad('IREGION', GRID_ARRAY_FORMAT_KEYWORDS, self._iregion),
            PropertyToLoad('PVMULT', GRID_ARRAY_FORMAT_KEYWORDS, self.__pvmult),
            PropertyToLoad('CORP', ['VALUE'], self.__corp),
            PropertyToLoad('IEQUIL', GRID_ARRAY_FORMAT_KEYWORDS, self.__iequil),
            PropertyToLoad('IPVT', GRID_ARRAY_FORMAT_KEYWORDS, self.__ipvt),
            PropertyToLoad('IWATER', GRID_ARRAY_FORMAT_KEYWORDS, self.__iwater),
            PropertyToLoad('IRELPM', GRID_ARRAY_FORMAT_KEYWORDS, self.__irelpm),
            PropertyToLoad('IROCK', GRID_ARRAY_FORMAT_KEYWORDS, self.__irock),
            PropertyToLoad('ITRAN', GRID_ARRAY_FORMAT_KEYWORDS, self.__itran),
            PropertyToLoad('LIVECELL', GRID_ARRAY_FORMAT_KEYWORDS, self.__livecell),
            PropertyToLoad('WORKA1', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka1),
            PropertyToLoad('WORKA2', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka2),
            PropertyToLoad('WORKA3', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka3),
            PropertyToLoad('WORKA4', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka4),
            PropertyToLoad('WORKA5', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka5),
            PropertyToLoad('WORKA6', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka6),
            PropertyToLoad('WORKA7', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka7),
            PropertyToLoad('WORKA8', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka8),
            PropertyToLoad('WORKA9', GRID_ARRAY_FORMAT_KEYWORDS, self.__worka9),
            PropertyToLoad('DX', GRID_ARRAY_FORMAT_KEYWORDS, self.__dx),
            PropertyToLoad('DY', GRID_ARRAY_FORMAT_KEYWORDS, self.__dy),
            PropertyToLoad('DZ', GRID_ARRAY_FORMAT_KEYWORDS, self.__dz),
            PropertyToLoad('DEPTH', GRID_ARRAY_FORMAT_KEYWORDS, self.__depth),
            PropertyToLoad('MDEPTH', GRID_ARRAY_FORMAT_KEYWORDS, self.__mdepth),
            PropertyToLoad('DZNET', GRID_ARRAY_FORMAT_KEYWORDS, self.__dznet),
            PropertyToLoad('COMPR', GRID_ARRAY_FORMAT_KEYWORDS, self.__compr),
            PropertyToLoad('CR', GRID_ARRAY_FORMAT_KEYWORDS, self.__compr),
            PropertyToLoad('ICOARS', GRID_ARRAY_FORMAT_KEYWORDS, self.__icoars),
            PropertyToLoad('IALPHAF', GRID_ARRAY_FORMAT_KEYWORDS, self.__ialphaf),
            PropertyToLoad('IPOLYMER', GRID_ARRAY_FORMAT_KEYWORDS, self.__ipolymer),
            PropertyToLoad('IADSORPTION', GRID_ARRAY_FORMAT_KEYWORDS, self.__iadsorption),
            PropertyToLoad('ITRACER', GRID_ARRAY_FORMAT_KEYWORDS, self.__itracer),
            PropertyToLoad('IGRID', GRID_ARRAY_FORMAT_KEYWORDS, self.__igrid),
            PropertyToLoad('ISECTOR', GRID_ARRAY_FORMAT_KEYWORDS, self.__isector),
            PropertyToLoad('SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__swl),
            PropertyToLoad('SWR', GRID_ARRAY_FORMAT_KEYWORDS, self.__swr),
            PropertyToLoad('SWU', GRID_ARRAY_FORMAT_KEYWORDS, self.__swu),
            PropertyToLoad('SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgl),
            PropertyToLoad('SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgr),
            PropertyToLoad('SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgu),
            PropertyToLoad('SWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__swro),
            PropertyToLoad('SWRO_LS', GRID_ARRAY_FORMAT_KEYWORDS, self.__swro_ls),
            PropertyToLoad('SGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgro),
            PropertyToLoad('SGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgrw),
            PropertyToLoad('KRW_SWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swro),
            PropertyToLoad('KRWRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swro),
            PropertyToLoad('KRWS_LS', GRID_ARRAY_FORMAT_KEYWORDS, self.__krws_ls),
            PropertyToLoad('KRW_SWU', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_swu),
            PropertyToLoad('KRG_SGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgro),
            PropertyToLoad('KRGRO', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgro),
            PropertyToLoad('KRG_SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgu),
            PropertyToLoad('KRG_SGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgrw),
            PropertyToLoad('KRGRW', GRID_ARRAY_FORMAT_KEYWORDS, self.__krg_sgrw),
            PropertyToLoad('KRO_SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swl),
            PropertyToLoad('KROLW', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swl),
            PropertyToLoad('KRO_SWR', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_swr),
            PropertyToLoad('KRO_SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_sgl),
            PropertyToLoad('KRO_SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__kro_sgr),
            PropertyToLoad('KRW_SGL', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_sgl),
            PropertyToLoad('KRW_SGR', GRID_ARRAY_FORMAT_KEYWORDS, self.__krw_sgr),
            PropertyToLoad('SGTR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sgtr),
            PropertyToLoad('SOTR', GRID_ARRAY_FORMAT_KEYWORDS, self.__sotr),
            PropertyToLoad('SWLPC', GRID_ARRAY_FORMAT_KEYWORDS, self.__swlpc),
            PropertyToLoad('SGLPC', GRID_ARRAY_FORMAT_KEYWORDS, self.__sglpc),
            PropertyToLoad('PCW_SWL', GRID_ARRAY_FORMAT_KEYWORDS, self.__pcw_swl),
            PropertyToLoad('PCG_SGU', GRID_ARRAY_FORMAT_KEYWORDS, self.__pcg_sgu),
            PropertyToLoad('CHLORIDE', GRID_ARRAY_FORMAT_KEYWORDS, self.__chloride),
            PropertyToLoad('CALCIUM', GRID_ARRAY_FORMAT_KEYWORDS, self.__calcium),
            PropertyToLoad('SALINITY', GRID_ARRAY_FORMAT_KEYWORDS, self.__salinity),
            PropertyToLoad('SAL', GRID_ARRAY_FORMAT_KEYWORDS, self.__salinity),
            PropertyToLoad('API', GRID_ARRAY_FORMAT_KEYWORDS, self.__api),
            PropertyToLoad('TMX', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmx),
            PropertyToLoad('TMY', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmy),
            PropertyToLoad('TMZ', GRID_ARRAY_FORMAT_KEYWORDS, self.__tmz),
            PropertyToLoad('MULTBV', GRID_ARRAY_FORMAT_KEYWORDS, self.__multbv),
            PropertyToLoad('PV', GRID_ARRAY_FORMAT_KEYWORDS, self.__pv)
        ]

        possible_properties = [prop.token for prop in properties_to_load]

        ignore_line = False
        array_name = 'ROOT'
        property_dict: dict = {}
        unit_system = self.__model_unit_system

        for idx, (original_line_location, line) in enumerate(file_as_list_with_original_line_numbers):

            # Load in the basic properties
            line_start_token = nfo.get_next_value(0, [line])

            if line_start_token is None:
                continue

            # Confirm we aren't looking at an area that has been commented out. If we are, continue to the next line.
            if line_start_token.upper() == 'NOSKIP':
                ignore_line = False

            if line_start_token.upper() == 'SKIP':
                ignore_line = True

            if ignore_line:
                continue

            nfo.check_property_in_line(line, property_dict, file_as_list)
            unit_system = property_dict.get('UNIT_SYSTEM', self.__model_unit_system)

            if nfo.check_token('ARRAYS', line):
                temp_array_name = fo.get_token_value('ARRAYS', line, [line])
                if temp_array_name is None:
                    array_name = 'ROOT'
                else:
                    array_name = temp_array_name

            if line_start_token is not None and line_start_token.upper() in possible_properties:
                token_property = properties_to_load[possible_properties.index(line_start_token)]
                for modifier in token_property.modifiers:
                    # execute the load
                    if array_name == 'ROOT':
                        StructuredGridOperations.load_token_value_if_present(
                            token_property.token, modifier, token_property.property, line, file_as_list, idx,
                            grid_nexus_file=self.__grid_nexus_file, ignore_values=['INCLUDE', 'NOLIST'],
                            original_line_location=original_line_location)
                    else:
                        # get the LGR object to add the grid array to:
                        lgr = self.lgrs.get(array_name)
                        attribute_name = self.keyword_mapping()[token_property.token][0]
                        grid_array_def_to_modify = getattr(lgr, attribute_name)
                        StructuredGridOperations.load_token_value_if_present(
                            token_property.token, modifier, grid_array_def_to_modify, line, file_as_list, idx,
                            grid_nexus_file=self.__grid_nexus_file, ignore_values=['INCLUDE', 'NOLIST'],
                            original_line_location=original_line_location)

            # Load in grid dimensions
            if nfo.check_token('NX', line):
                # Check that the format of the grid is NX followed by NY followed by NZ
                current_line = file_as_list[file_as_list.index(line)]
                remaining_line = current_line[current_line.index('NX') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NY':
                    continue
                remaining_line = remaining_line[remaining_line.index('NY') + 2:]
                if nfo.get_next_value(0, [remaining_line], remaining_line) != 'NZ':
                    continue

                # Avoid loading in a comment
                if "!" in line and line.index("!") < line.index('NX'):
                    continue
                next_line = file_as_list[file_as_list.index(line) + 1]
                first_value, next_line = move_next_value(next_line)
                second_value, next_line = move_next_value(next_line)
                third_value, next_line = move_next_value(next_line)

                self._range_x = int(first_value)
                self._range_y = int(second_value)
                self._range_z = int(third_value)

        # load the overs:
        if fo.value_in_file('OVER', file_as_list):
            self.__overs = NexusGrid.load_nexus_overs(file_as_list)
        if fo.value_in_file('FTRANS', file_as_list):
            self.__ftrans = NexusGrid.load_nexus_ftrans(file_as_list, unit_system)
        if fo.value_in_file('TOVER', file_as_list):
            self.__tovers = NexusGrid.load_nexus_tovers(file_as_list)

        self._grid_properties_loaded = True

    @classmethod
    def load_structured_grid_file(cls: type[NexusGrid], structured_grid_file: File,
                                  model_unit_system: UnitSystem, lazy_loading: bool = True,
                                  ) -> NexusGrid:
        """Loads in a structured grid file with all grid properties, and the array functions defined with 'FUNCTION'.

        Other grid modifiers are currently not supported.

        Args:
            structured_grid_file (NexusFile): the NexusFile representation of a structured grid file for converting \
                into a structured grid file class
            model_unit_system (UnitSystem): the default unit system to use for the model
            lazy_loading (bool): If set to True, parts of the grid will only be loaded in when requested via \
                properties on the object.

        Raises:
            AttributeError: if no value is found for the structured grid file path
            ValueError: if when loading the grid no values can be found for the NX NY NZ line.

        """
        if structured_grid_file.location is None:
            raise ValueError(f"No file path given or found for structured grid file path. \
                Instead got {structured_grid_file.location}")

        if not isinstance(structured_grid_file, NexusFile):
            raise ValueError(f"Cannot load file of type {type(structured_grid_file)}.")

        loaded_structured_grid_file = cls(grid_nexus_file=structured_grid_file, model_unit_system=model_unit_system)

        if not lazy_loading:
            loaded_structured_grid_file.load_grid_properties_if_not_loaded()
            loaded_structured_grid_file.load_faults()
            loaded_structured_grid_file.load_array_functions()

        return loaded_structured_grid_file

    @staticmethod
    def update_structured_grid_file(grid_dict: dict[str, GridArrayDefinition | int], model: NexusSimulator) -> None:
        """Save values passed from the front end to the structured grid file and update the class.

        Args:
            grid_dict (dict[str, Union[VariableEntry, int]]): dictionary containing grid properties to be replaced
            model (NexusSimulator): an instance of a NexusSimulator object
        Raises:
            ValueError: If no structured grid file is in the instance of the Simulator class
        """
        # Convert the dictionary back to a class, and update the properties on our class
        structured_grid = model.grid
        if structured_grid is None or model.model_files.structured_grid_file is None:
            raise ValueError("Model does not contain a structured grid")
        original_structured_grid_file = copy.deepcopy(structured_grid)

        # replace the structured grid with a new object with an updated dictionary
        structured_grid.update_properties_from_dict(grid_dict)

        # change it in the text file for nexus:
        grid_file_path = model.model_files.structured_grid_file.location
        if grid_file_path is None:
            raise ValueError("No path found for structured grid file path.")
        structured_grid_contents = nfo.load_file_as_list(grid_file_path)
        # Get the existing file as a list
        if structured_grid_contents is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        # Update each value in the file
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.netgrs,
                                               structured_grid.netgrs, 'NETGRS')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.porosity,
                                               structured_grid.porosity, 'POROSITY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.sw,
                                               structured_grid.sw, 'SW')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kx,
                                               structured_grid.kx, 'KX')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.ky,
                                               structured_grid.ky, 'KY')
        StructuredGridOperations.replace_value(structured_grid_contents, original_structured_grid_file.kz,
                                               structured_grid.kz, 'KZ')

        # Save the new file contents
        new_file_str = "".join(structured_grid_contents)
        with open(grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    def load_array_functions(self) -> None:
        """Loads collection of array function defined in the nexus grid file."""
        # for function arrays we need the expanded file contents without includes
        if self.__grid_nexus_file is None or self.__grid_nexus_file.get_flat_list_str_file is None:
            raise ValueError("Cannot load array functions as grid file cannot not found")
        file_contents = self.__grid_nexus_file.get_flat_list_str_file
        self.__array_functions_list = afo.collect_all_function_blocks(file_contents)
        self.__grid_array_functions = afo.create_grid_array_function_objects(self.__array_functions_list)
        self.__array_functions_df = afo.summarize_model_functions(self.__array_functions_list)
        self.__array_functions_loaded = True

    def get_array_functions_list(self) -> Optional[list[list[str]]]:
        """Returns the grid array functions as a list of function lines."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_list

    def get_array_functions_df(self) -> Optional[pd.DataFrame]:
        """Returns the grid array functions as a dataframe."""
        if not self.__array_functions_loaded:
            self.load_array_functions()
        return self.__array_functions_df

    def load_faults(self) -> None:
        """Function to read faults in Nexus grid file defined using MULT and FNAME keywords."""
        file_content_as_list = self.__grid_file_contents
        if file_content_as_list is None:
            raise ValueError('Grid file contents have not been loaded')
        df = load_nexus_fault_mult_table_from_list(file_content_as_list)

        if not df.empty:
            # Ensure resulting dataframe has uppercase column names
            df.columns = [col.upper() for col in df.columns]

            # Check if any multfls have been used in grid file and update fault trans multipliers accordingly
            f_names = df['NAME'].unique()
            f_mults = [1.] * len(f_names)
            mult_dict = dict(zip(f_names, f_mults))
            for line in file_content_as_list:
                if nfo.check_token('MULTFL', line):
                    fname = str(nfo.get_expected_token_value(
                        'MULTFL', line, file_content_as_list,
                        custom_message=f'{line} does not have a fault name following MULTFL'))
                    if fname in df['NAME'].unique():
                        tmult = float(str(nfo.get_expected_token_value(
                            fname, line, file_content_as_list,
                            custom_message=f'MULTFL {fname} does not have a numerical tmult value')))
                        mult_dict[fname] *= tmult
            mult_df = pd.DataFrame.from_dict(mult_dict, orient='index').reset_index()
            mult_df.columns = ['NAME', 'TMULT']
            new_df = df.merge(mult_df, how='left', on='NAME', validate='many_to_one')
            new_df['MULT'] = new_df['MULT'] * new_df['TMULT']
            self.__faults_df = new_df.drop(['TMULT'], axis=1)
        self.__grid_faults_loaded = True

    def get_faults_df(self) -> Optional[pd.DataFrame]:
        """Returns the fault definition and transmissility multiplier information as a dataframe."""
        self.load_grid_properties_if_not_loaded()
        if not self.__grid_faults_loaded:
            self.load_faults()
        return self.__faults_df

    def load_multir(self) -> None:
        """Function to read MULTIR in Nexus grid file."""
        file_content_as_list = self.__grid_file_contents
        if file_content_as_list is None:
            raise ValueError('Grid file contents have not been loaded')
        multir_list = self.load_nexus_multir_table_from_list(file_content_as_list)
        self.__multir = multir_list
        self.__grid_multir_loaded = True

    @staticmethod
    def load_nexus_multir_table_from_list(file_content_as_list: list[str]) -> list[NexusMultir]:
        """Function to read MULTIR from a file represented as a list."""
        start_idx = -1
        end_idx = -1

        valid_end_tokens = VALID_NEXUS_KEYWORDS
        skip_tokens = ['X', 'Y', 'Z', 'XYZ', 'ALL', 'STD', 'NONSTD']
        valid_end_tokens = [x for x in valid_end_tokens if x not in skip_tokens]
        multir_lists: list[NexusMultir] = []

        for idx, line in enumerate(file_content_as_list):
            if nfo.nexus_token_found(line, valid_end_tokens):
                end_idx = idx
            if idx == len(file_content_as_list) - 1:
                end_idx = idx + 1

            if 0 < start_idx < end_idx - 1:
                for multir_line in file_content_as_list[start_idx:end_idx]:
                    new_multir = NexusGrid.__extract_multir_tableline(multir_line)
                    if new_multir is not None:
                        multir_lists.append(new_multir)
                start_idx = -1
                end_idx = -1

            if nfo.check_token('MULTIR', line):
                start_idx = idx + 1
                continue
        # if no MULTIR table is found, return an empty list
        return multir_lists

    @staticmethod
    def __extract_multir_tableline(line: str) -> None | NexusMultir:
        """Takes a single line in a file and extracts a Multir object from it."""
        stored_values = nfo.split_line(line)
        if not stored_values:
            return None
        region_1_str, region_2_str, tmult_str = stored_values[0:3]
        region_1 = int(region_1_str)
        region_2 = int(region_2_str)
        tmult = float(tmult_str)

        direction = ''
        for ele in stored_values:
            if 'X' in ele:
                direction += 'X'
            if 'Y' in ele:
                direction += 'Y'
            if 'Z' in ele:
                direction += 'Z'

        if direction == '':
            # default to all directions
            direction = 'XYZ'

        standard_connections = True
        non_standard_connections = True
        if 'STD' in stored_values and 'NONSTD' not in stored_values:
            standard_connections = True
            non_standard_connections = False
        if 'NONSTD' in stored_values and 'STD' not in stored_values:
            standard_connections = False
            non_standard_connections = True
        if 'ALL' in stored_values or ('STD' not in stored_values and 'NONSTD' not in stored_values):
            standard_connections = True
            non_standard_connections = True
        return NexusMultir(region_1=region_1, region_2=region_2, tmult=tmult, directions=direction,
                           std_connections=standard_connections, non_std_connections=non_standard_connections)

    def get_multir(self) -> list[NexusMultir]:
        """Returns the MULTIR information as a list of multir objects."""
        self.load_grid_properties_if_not_loaded()
        if not self.__grid_multir_loaded:
            self.load_multir()
        return self.__multir if self.__multir is not None else []

    @staticmethod
    def __keyword_in_include_file_warning(var_entry_obj: GridArrayDefinition) -> None:

        if var_entry_obj.keyword_in_include_file is True:
            warnings.warn('Grid array keyword in include file. This is not recommended simulation practice.')
        else:
            return None

    @property
    def array_functions(self) -> Optional[list[NexusGridArrayFunction]]:
        """Returns a list of the array functions defined in the structured grid file."""
        self.load_grid_properties_if_not_loaded()
        if self.__grid_array_functions is None:
            self.load_array_functions()
        return self.__grid_array_functions

    @property
    def corp(self) -> GridArrayDefinition:
        """Returns corp grid property.
        Ensures grid properties are loaded and returns value of 'corp'.
        """
        self.load_grid_properties_if_not_loaded()
        NexusGrid.__keyword_in_include_file_warning(self.__corp)

        return self.__corp

    @property
    def iequil(self) -> GridArrayDefinition:
        """Returns iequil grid property.
        Ensures grid properties are loaded and returns value of 'iequil'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__iequil

    @property
    def ipvt(self) -> GridArrayDefinition:
        """Returns ipvt grid property.
        Ensures grid properties are loaded and returns value of 'ipvt'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__ipvt

    @property
    def iwater(self) -> GridArrayDefinition:
        """Returns iwater grid property.
        Ensures grid properties are loaded and returns value of 'iwater'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__iwater

    @property
    def irelpm(self) -> GridArrayDefinition:
        """Returns irelpm grid property.
        Ensures grid properties are loaded and returns value of 'irelpm'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__irelpm

    @property
    def irock(self) -> GridArrayDefinition:
        """Returns irock grid property.
        Ensures grid properties are loaded and returns value of 'irock'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__irock

    @property
    def itran(self) -> GridArrayDefinition:
        """Returns itran grid property.
        Ensures grid properties are loaded and returns value of 'itran'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__itran

    @property
    def iregion(self) -> dict[str, GridArrayDefinition]:
        """Returns iregion grid property.
        Ensures grid properties are loaded and returns value of 'iregion'.
        """
        self.load_grid_properties_if_not_loaded()
        return self._iregion

    @property
    def livecell(self) -> GridArrayDefinition:
        """Returns livecell grid property.
        Ensures grid properties are loaded and returns value of 'livecell'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__livecell

    @property
    def pvmult(self) -> GridArrayDefinition:
        """Returns pvmult grid property.
        Ensures grid properties are loaded and returns value of 'pv_mult'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__pvmult

    @property
    def worka1(self) -> GridArrayDefinition:
        """Returns worka1 grid property.
        Ensures grid properties are loaded and returns value of 'worka1'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka1

    @property
    def worka2(self) -> GridArrayDefinition:
        """Returns worka2 grid property.
        Ensures grid properties are loaded and returns value of 'worka2'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka2

    @property
    def worka3(self) -> GridArrayDefinition:
        """Returns worka3 grid property.
        Ensures grid properties are loaded and returns value of 'worka3'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka3

    @property
    def worka4(self) -> GridArrayDefinition:
        """Returns worka4 grid property.
        Ensures grid properties are loaded and returns value of 'worka4'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka4

    @property
    def worka5(self) -> GridArrayDefinition:
        """Returns worka5 grid property.
        Ensures grid properties are loaded and returns value of 'worka5'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka5

    @property
    def worka6(self) -> GridArrayDefinition:
        """Returns worka6 grid property.
        Ensures grid properties are loaded and returns value of 'worka6'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka6

    @property
    def worka7(self) -> GridArrayDefinition:
        """Returns worka7 grid property.
        Ensures grid properties are loaded and returns value of 'worka7'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka7

    @property
    def worka8(self) -> GridArrayDefinition:
        """Returns worka8 grid property.
        Ensures grid properties are loaded and returns value of 'worka8'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka8

    @property
    def worka9(self) -> GridArrayDefinition:
        """Returns worka9 grid property.
        Ensures grid properties are loaded and returns value of 'worka9'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__worka9

    @property
    def dx(self) -> GridArrayDefinition:
        """Returns dx grid property.
        Ensures grid properties are loaded and returns value of 'dx'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__dx

    @property
    def dy(self) -> GridArrayDefinition:
        """Returns dy grid property.
        Ensures grid properties are loaded and returns value of 'dy'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__dy

    @property
    def dz(self) -> GridArrayDefinition:
        """Returns dz grid property.
        Ensures grid properties are loaded and returns value of 'dz'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__dz

    @property
    def depth(self) -> GridArrayDefinition:
        """Returns depth grid property.
        Ensures grid properties are loaded and returns value of 'depth'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__depth

    @property
    def mdepth(self) -> GridArrayDefinition:
        """Returns mdepth grid property.
        Ensures grid properties are loaded and returns value of 'mdepth'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__mdepth

    @property
    def dznet(self) -> GridArrayDefinition:
        """Returns dznet grid property.
        Ensures grid properties are loaded and returns value of 'dznet'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__dznet

    @property
    def compr(self) -> GridArrayDefinition:
        """Returns compr grid property.
        Ensures grid properties are loaded and returns value of 'compr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__compr

    @property
    def icoars(self) -> GridArrayDefinition:
        """Returns icoars grid property.
        Ensures grid properties are loaded and returns value of 'icoars'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__icoars

    @property
    def ialphaf(self) -> GridArrayDefinition:
        """Returns ialphaf grid property.
        Ensures grid properties are loaded and returns value of 'ialphaf'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__ialphaf

    @property
    def ipolymer(self) -> GridArrayDefinition:
        """Returns ipolymer grid property.
        Ensures grid properties are loaded and returns value of 'ipolymer'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__ipolymer

    @property
    def iadsorption(self) -> GridArrayDefinition:
        """Returns iadsorption grid property.
        Ensures grid properties are loaded and returns value of 'iadsorption'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__iadsorption

    @property
    def itracer(self) -> GridArrayDefinition:
        """Returns itracer grid property.
        Ensures grid properties are loaded and returns value of 'itracer'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__itracer

    @property
    def igrid(self) -> GridArrayDefinition:
        """Returns igrid grid property.
        Ensures grid properties are loaded and returns value of 'igrid'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__igrid

    @property
    def isector(self) -> GridArrayDefinition:
        """Returns isector grid property.
        Ensures grid properties are loaded and returns value of 'isector'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__isector

    @property
    def swl(self) -> GridArrayDefinition:
        """Returns swl grid property.
        Ensures grid properties are loaded and returns value of 'swl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swl

    @property
    def swr(self) -> GridArrayDefinition:
        """Returns swr grid property.
        Ensures grid properties are loaded and returns value of 'swr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swr

    @property
    def swu(self) -> GridArrayDefinition:
        """Returns swu grid property.
        Ensures grid properties are loaded and returns value of 'swu'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swu

    @property
    def sgl(self) -> GridArrayDefinition:
        """Returns sgl grid property.
        Ensures grid properties are loaded and returns value of 'sgl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgl

    @property
    def sgr(self) -> GridArrayDefinition:
        """Returns sgr grid property.
        Ensures grid properties are loaded and returns value of 'sgr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgr

    @property
    def sgu(self) -> GridArrayDefinition:
        """Returns sgu grid property.
        Ensures grid properties are loaded and returns value of 'sgu'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgu

    @property
    def swro(self) -> GridArrayDefinition:
        """Returns swro grid property.
        Ensures grid properties are loaded and returns value of 'swro'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swro

    @property
    def swro_ls(self) -> GridArrayDefinition:
        """Returns swro_ls grid property.
        Ensures grid properties are loaded and returns value of 'swro_ls'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swro_ls

    @property
    def sgro(self) -> GridArrayDefinition:
        """Returns sgro grid property.
        Ensures grid properties are loaded and returns value of 'sgro'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgro

    @property
    def sgrw(self) -> GridArrayDefinition:
        """Returns sgrw grid property.
        Ensures grid properties are loaded and returns value of 'sgrw'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgrw

    @property
    def krw_swro(self) -> GridArrayDefinition:
        """Returns krw_swro grid property.
        Ensures grid properties are loaded and returns value of 'krw_swro'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krw_swro

    @property
    def krws_ls(self) -> GridArrayDefinition:
        """Returns krws_ls grid property.
        Ensures grid properties are loaded and returns value of 'krws_ls'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krws_ls

    @property
    def krw_swu(self) -> GridArrayDefinition:
        """Returns krw_swu grid property.
        Ensures grid properties are loaded and returns value of 'krw_swu'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krw_swu

    @property
    def krg_sgro(self) -> GridArrayDefinition:
        """Returns kro_sgro grid property.
        Ensures grid properties are loaded and returns value of 'krg_sgro'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgro

    @property
    def krg_sgu(self) -> GridArrayDefinition:
        """Returns krg_sgu grid property.
        Ensures grid properties are loaded and returns value of 'krg_sgu'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgu

    @property
    def krg_sgrw(self) -> GridArrayDefinition:
        """Returns kro_sgrw grid property.
        Ensures grid properties are loaded and returns value of 'krg_sgrw'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krg_sgrw

    @property
    def kro_swl(self) -> GridArrayDefinition:
        """Returns kro_swl grid property.
        Ensures grid properties are loaded and returns value of 'kro_swl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kro_swl

    @property
    def kro_swr(self) -> GridArrayDefinition:
        """Returns kro_swr grid property.
        Ensures grid properties are loaded and returns value of 'kro_swr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kro_swr

    @property
    def kro_sgl(self) -> GridArrayDefinition:
        """Returns kro_sgl grid property.
        Ensures grid properties are loaded and returns value of 'kro_sgl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kro_sgl

    @property
    def kro_sgr(self) -> GridArrayDefinition:
        """Returns kro_sgr grid property.
        Ensures grid properties are loaded and returns value of 'kro_sgr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kro_sgr

    @property
    def krw_sgl(self) -> GridArrayDefinition:
        """Returns krw_sgl grid property.
        Ensures grid properties are loaded and returns value of 'krw_sgl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krw_sgl

    @property
    def krw_sgr(self) -> GridArrayDefinition:
        """Returns krw_sgr grid property.
        Ensures grid properties are loaded and returns value of 'krw_sgr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__krw_sgr

    @property
    def sgtr(self) -> GridArrayDefinition:
        """"Returns sgtr grid property.
        Ensures gir properties are loaded and returns value of 'sgtr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sgtr

    @property
    def sotr(self) -> GridArrayDefinition:
        """"Returns sotr grid property.
        Ensures grid properties are loaded and returns value of 'sotr'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sotr

    @property
    def swlpc(self) -> GridArrayDefinition:
        """Returns swlpc grid property.
        Ensures grid properties are loaded and returns value of 'swlpc'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__swlpc

    @property
    def sglpc(self) -> GridArrayDefinition:
        """Returns sglpc property.
        Ensures grid properties are loaded and returns value of 'sglpc'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__sglpc

    @property
    def pcw_swl(self) -> GridArrayDefinition:
        """Returns pcw_swl property.
        Ensures grid properties are loaded and returns value of 'pcw_swl'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__pcw_swl

    @property
    def pcg_sgu(self) -> GridArrayDefinition:
        """Returns pcg_sgu grid proprety.
        Ensures grid properties are loaded and returns value of 'pcg_sgu'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__pcg_sgu

    @property
    def chloride(self) -> GridArrayDefinition:
        """Returns chloride grid property.
        Ensures grid properties are loaded and returns value of 'chloride'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__chloride

    @property
    def calcium(self) -> GridArrayDefinition:
        """Returns calcium grid property.
        Ensures grid properties are loaded and returns value of 'calcium'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__calcium

    @property
    def salinity(self) -> GridArrayDefinition:
        """Returns grid salinity.
        Ensures grid properties are loaded and returns the value 'salinity'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__salinity

    @property
    def api(self) -> GridArrayDefinition:
        """Returns grid api.
        Ensures grid properties are loaded and returns value of 'api'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__api

    @property
    def tmx(self) -> GridArrayDefinition:
        """Returns gird tmx.
        Ensures grid properties are loaded and returns value of 'tmx'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__tmx

    @property
    def tmy(self) -> GridArrayDefinition:
        """Returns grid tmy.
        Ensures grid properties are loaded and returns value of 'tmy'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__tmy

    @property
    def tmz(self) -> GridArrayDefinition:
        """Returns grid tmz.
        Ensures gird properties are loaded and returns value if 'tmz'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__tmz

    @property
    def multbv(self) -> GridArrayDefinition:
        """Returns grid multvb.
        Ensures grid properties are loaded and returns value of 'multvb'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__multbv

    @property
    def pv(self) -> GridArrayDefinition:
        """Returns grid pv.
        Ensures grid properties are loaded and returns value of 'pv'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__pv

    @property
    def kxeff(self) -> GridArrayDefinition:
        """Returns the kxeff grid property.
        Ensures grid properties are loaded and returns value of 'kxeff'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kxeff

    @property
    def kyeff(self) -> GridArrayDefinition:
        """Returns the kyeff grid property.
        Ensures grid properties are loaded and returns value of 'kyeff'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kyeff

    @property
    def kzeff(self) -> GridArrayDefinition:
        """Returns the kzeff grid property.
        Ensures grid properties are loaded and returns value of 'kzeff'.
        """
        self.load_grid_properties_if_not_loaded()
        return self.__kzeff

    @property
    def multir(self) -> list[NexusMultir]:
        """Returns the MULTIR table as a list of multir objects."""
        return self.get_multir()

    @property
    def lgrs(self) -> NexusLGRs:
        """Returns the LGR object which contains a list of the LGRs in the nexus grid."""
        return self.__lgrs

    @staticmethod
    def load_nexus_overs(file_content_as_list: list[str]) -> list[NexusOver]:
        """Function to read in OVER tables from a file.

        Args:
            file_content_as_list (list[str]): list of strings representing the file contents.

        Returns:
            list[NexusOver]: list of NexusOver objects representing the OVER table.
        """
        ignore_list = ['OVER', 'TX', 'TY', 'TZ', 'PV', 'PVF', 'TXF', 'TYF', 'TZF', 'GRID', 'FNAME', 'GE', 'LE', 'ROOT']
        valid_end_tokens = [x for x in VALID_NEXUS_KEYWORDS if x not in ignore_list]
        overs_list: list[NexusOver] = []
        reading_over = False
        grid = 'ROOT'
        fname = None
        arrays: list[str] = []
        potential_operators: Final = ['+', '-', '*', '/', '=']
        threshold_value = None
        for line in file_content_as_list:
            if nfo.nexus_token_found(line, valid_end_tokens):
                reading_over = False
                grid = 'ROOT'
                fname = None
                arrays = []
                continue
            if reading_over:
                split_line = nfo.split_line(line)
                if nfo.check_token('GRID', line):
                    grid = nfo.get_expected_token_value('GRID', line, file_content_as_list)
                if nfo.check_token('FNAME', line):
                    fname = nfo.get_expected_token_value('FNAME', line, file_content_as_list)
                if len(split_line) > 6:
                    i1, i2, j1, j2, k1, k2 = (int(x) for x in split_line[0:6])
                    # cut out the ranges
                    split_line = split_line[6:]
                    for array in arrays:
                        operator_value = split_line[0]
                        operator_matches = [x for x in potential_operators if x == operator_value[0]]
                        if not operator_matches and ('GE' in split_line or 'LE' in split_line):
                            # if the operator is not found then it is GE or LE
                            value = float(split_line[0])
                            operator = split_line[1]
                            threshold_value = float(split_line[2])
                            split_line_position = 3
                        elif operator_matches:
                            operator = operator_matches[0]
                            # remove the operator and the remaining string is the value
                            trimmed_value = split_line[0].replace(operator, '')
                            if trimmed_value == '':
                                # then the value is in the next element of split_line
                                value = float(split_line[1])
                                split_line_position = 2
                            else:
                                value = float(split_line[0][1:])
                                split_line_position = 1
                        else:
                            # no operator match and not GE or LE then it is implicitly '*'
                            value = float(split_line[0])
                            operator = '*'
                            split_line_position = 1
                        overs_list.append(NexusOver(array=array, grid=grid, fault_name=fname,
                                                    i1=i1, i2=i2, j1=j1, j2=j2, k1=k1, k2=k2, operator=operator,
                                                    value=value, threshold=threshold_value))
                        split_line = split_line[split_line_position:]
                        threshold_value = None

            if nfo.check_token('OVER', line):
                reading_over = True
                over_split_line = nfo.split_line(line)
                arrays = over_split_line[over_split_line.index('OVER') + 1:]

        return overs_list

    @property
    def overs(self) -> list[NexusOver]:
        """Returns the OVER table as a list of NexusOver objects."""
        if not self._grid_properties_loaded:
            self.load_grid_properties_if_not_loaded()
        return self.__overs

    @property
    def tovers(self) -> list[NexusTOver]:
        """Returns the TOVER table as a list of NexusTOver objects."""
        if not self._grid_properties_loaded:
            self.load_grid_properties_if_not_loaded()
        return self.__tovers

    @staticmethod
    def load_nexus_tovers(file_content_as_list: list[str]) -> list[NexusTOver]:
        """Loads the Nexus TOVER tables to a list of objects.

        Args:
        file_content_as_list (list[str]): list of strings representing the file contents.

        Returns:
            list[NexusOver]: list of NexusTOver objects representing the TOVER table.
        """
        ignore_list = ['OVER', 'TX', 'TY', 'TZ', 'GRID', 'ROOT', 'TOVER',
                       'TX+', 'TX-', 'TZ+', 'TZ-', 'TY+', 'TY-',
                       'TXF+', 'TXF-', 'TYF+', 'TYF-', 'TZF+', 'TZF-', 'INCLUDE',
                       'ADD', 'SUB', 'DIV', 'EQ', 'MULT']
        valid_end_tokens = [x for x in VALID_NEXUS_KEYWORDS if x not in ignore_list]
        tovers_list: list[NexusTOver] = []
        reading_tover = False
        grid = 'ROOT'
        array = ''
        potential_operators: Final = ['ADD', 'SUB', 'DIV', 'MULT', 'EQ']
        i1, i2, j1, j2, k1, k2 = 0, 0, 0, 0, 0, 0
        operator = ''
        for i, line in enumerate(file_content_as_list):
            if nfo.check_token('TOVER', line.upper()):
                array = nfo.get_expected_token_value(token='TOVER', token_line=line,
                                                     file_list=file_content_as_list[i:])
                reading_tover = True

            if not reading_tover:
                continue
            if nfo.nexus_token_found(line, valid_end_tokens):
                # reset reading after another token found
                reading_tover = False
                grid = 'ROOT'
                array = ''
                operator = ''
                i1, i2, j1, j2, k1, k2 = 0, 0, 0, 0, 0, 0
                continue

            if any(nfo.check_token(x, line.upper()) for x in potential_operators):
                split_line = nfo.split_line(line)
                i1, i2, j1, j2, k1, k2 = (int(x) for x in split_line[0:6])
                operator = split_line[-1]
                continue
            if nfo.check_token('INCLUDE', line.upper()):
                include_file = nfo.get_expected_token_value(token='INCLUDE', token_line=line,
                                                            file_list=file_content_as_list[i:])
                new_tover = NexusTOver(i1=i1, i2=i2, j1=j1, j2=j2, k1=k1, k2=k2,
                                       include_file=include_file, array=array, grid=grid, operator=operator,
                                       value=0)
                tovers_list.append(new_tover)
                i1, i2, j1, j2, k1, k2 = 0, 0, 0, 0, 0, 0
                operator = ''

            elif i1 and i2 and j1 and j2 and k1 and k2 and operator and nfo.get_next_value(0, [line]):
                # not an include file, so it must be a value or array of values
                number_of_values = (i2 - i1 + 1) * (j2 - j1 + 1) * (k2 - k1 + 1)
                array_values = fo.get_multiple_expected_sequential_values(file_content_as_list[i:],
                                                                          number_tokens=number_of_values,
                                                                          ignore_values=[])
                new_tover = NexusTOver(i1=i1, i2=i2, j1=j1, j2=j2, k1=k1, k2=k2,
                                       include_file=None, array=array, grid=grid, operator=operator,
                                       value=0, array_values=[float(x) for x in array_values])
                i1, i2, j1, j2, k1, k2 = 0, 0, 0, 0, 0, 0
                operator = ''
                tovers_list.append(new_tover)

        return tovers_list

    @staticmethod
    def load_nexus_ftrans(file_content_as_list: list[str], unit_system: UnitSystem) -> list[NexusFtrans]:
        """Function to read in FTRANS tables from a file.

        Args:
            file_content_as_list (list[str]): list of strings representing the file contents.
            unit_system (UnitSystem): the unit system used in the grid file.

        Returns:
            list[NexusFtrans]: list of NexusFtrans objects representing the FTRANS table.
        """
        ignore_list = ['FTRANS', 'GRID', 'FNAME', 'ROOT']
        valid_end_tokens = [x for x in VALID_NEXUS_KEYWORDS if x not in ignore_list]
        ftrans_list: list[NexusFtrans] = []
        reading = False
        grid = 'ROOT'
        fname = None
        for line in file_content_as_list:
            if nfo.nexus_token_found(line, valid_end_tokens):
                reading = False
                grid = 'ROOT'
                fname = None
                continue

            if reading:
                split_line = nfo.split_line(line)
                if nfo.check_token('GRID', line):
                    grid = nfo.get_expected_token_value('GRID', line, file_content_as_list)
                if nfo.check_token('FNAME', line):
                    fname = nfo.get_expected_token_value('FNAME', line, file_content_as_list)
                if len(split_line) == 7:
                    i1, j1, k1, i2, j2, k2 = (int(x) for x in split_line[0:6])
                    value = float(split_line[-1])
                    # cut out the ranges
                    ftrans_list.append(NexusFtrans(grid=grid, fault_name=fname,
                                                   i1=i1, i2=i2, j1=j1, j2=j2, k1=k1, k2=k2,
                                                   value=value, unit_system=unit_system))

            if nfo.check_token('FTRANS', line):
                # reset the default values if another FTRANS call is found
                grid = 'ROOT'
                fname = None
                reading = True

        return ftrans_list

    @property
    def ftrans(self) -> list[NexusFtrans]:
        """Returns the OVER table as a list of NexusOver objects."""
        if not self._grid_properties_loaded:
            self.load_grid_properties_if_not_loaded()
        return self.__ftrans



"""Abstract base class for a node in a network."""
from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class Node(NetworkObject, ABC):
    type: Optional[str] = None
    depth: Optional[float] = None

    def __init__(self, properties_dict: Optional[dict[str, None | int | str | float]] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None, type: Optional[str] = None,
                 depth: Optional[float] = None) -> None:
        """Initialises the NexusNode class.

        Args:
            properties_dict (Optional[dict]): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            name (Optional[str]): The name of the object.
            type (Optional[str]): Well type.
            depth (Optional[float]): The depth of the wellhead.
        """

        self.type = type
        self.depth = depth

        super().__init__(properties_dict=properties_dict, date=date, date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=name)

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the constraint."""
        return NetworkUnits(self.unit_system)



from dataclasses import dataclass
from typing import Optional

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataModelBaseClasses.NetworkObject import NetworkObject
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Units.AttributeMappings.BaseUnitMapping import BaseUnitMapping


@dataclass(kw_only=True, repr=True)
class NexusAction(NetworkObject):
    """A class representing a single Nexus action."""

    action_time: float  # don't assume it is an int just yet
    change: ActivationChangeEnum
    connection: str

    def __init__(self, properties_dict: Optional[dict[str, None | int | str | float]] = None,
                 date: Optional[str] = None,
                 date_format: Optional[DateFormat] = None,
                 start_date: Optional[str] = None, unit_system: Optional[UnitSystem] = None) -> None:
        """Initializes the Nexus Action attributes."""

        # the fact that date is being explicitly set in the super call means that it would
        # technically be an unused parameter, but date is a requirement in the __init__ so we cannot
        # just get rid of it. So we implement a quick hack below.
        # This IF statement is necessary to pass the ruff check
        if date is None:
            pass

        # the action_time in the properties dict is the DATE
        action_time = ''  # expecting a float
        connection = ''

        if isinstance(properties_dict, dict):
            if isinstance(properties_dict['action_time'], float):
                action_time = str(properties_dict['action_time'])
            if isinstance(properties_dict['connection'], str):
                connection = properties_dict['connection']
        super().__init__(properties_dict=properties_dict, date=action_time,
                         date_format=date_format, start_date=start_date,
                         unit_system=unit_system, name=connection)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of keywords to attribute definitions."""
        keywords = {
            'ACTIONTIME': ('action_time', float),
            'ACTION': ('change', ActivationChangeEnum),
            'CONNECTION': ('connection', str),
        }
        return keywords

    @property
    def units(self) -> BaseUnitMapping:
        """Returns the attribute to unit map for the data object."""
        return BaseUnitMapping(unit_system=self.unit_system)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import numpy as np
import pandas as pd

from ResSimpy.FileOperations.File import File
from ResSimpy.FileOperations import file_operations as fo
from ResSimpy.Utils.grid_filtering_functions import grid_filter_file_as_list, filter_grid_array_definition


@dataclass
class GridArrayDefinition:
    """A class to define a grid array property.

    Args:
        modifier (Optional[str]): the modifier for the grid array property (e.g. CON, MULT, etc.)
        value (Optional[str]): the actual values for the grid array property in question. Can be an include file.
        mods (Optional[dict[str, pd.DataFrame]): if the grid array has an associated mod card we capture it.
        keyword_in_include_file (bool): an indicator to tell you if the grid array keyword was found in an inc file
        absolute_path (Optional[str]): the absolute path to the path if value is an include file.
    """
    modifier: Optional[str] = None
    value: Optional[str] = None
    mods: Optional[dict[str, pd.DataFrame]] = None
    keyword_in_include_file: bool = False
    absolute_path: Optional[str] = None
    array: Optional[np.ndarray] = None

    def load_grid_array_definition_to_file_as_list(self) -> list[str]:
        """Loads the grid array definition to a file as a list of strings."""
        path = self.absolute_path
        if path is None and self.value is not None:
            sanitised_value = "".join(fo.strip_file_of_comments(self.value.splitlines()))
            # If the value is a path, load that file in. Otherwise, assume it is a grid array value.
            if sanitised_value.upper().isupper():  # Fastest way to check if there are any letters in a string
                path = self.value
            else:
                return sanitised_value.splitlines()
        elif path is None:
            raise FileNotFoundError('No file path found in the grid array definition')
        return fo.load_file_as_list(path)

    def get_array_from_file(self, x_range: None | int = None, y_range: None | int = None,
                            z_range: None | int = None) -> np.ndarray:
        """Returns a 1D numpy array from the grid array definition."""
        if self.array is not None:
            return self.array
        file_as_list = self.load_grid_array_definition_to_file_as_list()
        self.array = self.grid_file_as_list_to_numpy_array(file_as_list, x_range, y_range, z_range)
        return self.array

    def filtered_grid_array_def_as_file(self) -> File:
        """Filters the grid array definition."""
        return filter_grid_array_definition(self)

    @staticmethod
    def grid_file_as_list_to_numpy_array(file_as_list: list[str], x_range: None | int, y_range: None | int,
                                         z_range: None | int) -> np.ndarray:
        """Converts a list of strings to a numpy array."""
        # ensure the list of strings is filtered of comments and non-float values
        new_file_as_list = grid_filter_file_as_list(file_as_list)

        array_size = -1
        if x_range is not None and y_range is not None and z_range is not None:
            array_size = x_range * y_range * z_range

        grid_array = np.fromstring(' '.join(new_file_as_list), sep=' ', count=array_size)

        return grid_array

    def _get_array_or_value(self) -> np.ndarray | float:
        """Returns the array or value based on the modifier."""
        if self.modifier == 'VALUE':
            return self.get_array_from_file()
        elif self.modifier == 'CON':
            if self.value is None:
                raise ValueError('No value found for grid array definition.')
            return float(self.value)
        else:
            raise NotImplementedError('Only VALUE and CON are modifiers supported for grid array definitions.')

    def min(self) -> float:
        """Returns the minimum value of the grid array property."""
        array_or_value = self._get_array_or_value()
        if isinstance(array_or_value, float):
            return array_or_value
        else:
            return array_or_value.min()

    def max(self) -> float:
        """Returns the maximum value of the grid array property."""
        array_or_value = self._get_array_or_value()
        if isinstance(array_or_value, float):
            return array_or_value
        else:
            return array_or_value.max()

    def get_array(self) -> np.ndarray:
        """Returns the array from the grid array definition."""
        return self.get_array_from_file()



"""This module contains the abstract base class for file manipulations for simulator files."""
from __future__ import annotations
import os
from uuid import uuid4, UUID
from dataclasses import dataclass, field
from typing import Optional, Sequence
import warnings
from ResSimpy.FileOperations.FileBase import FileBase
import ResSimpy.FileOperations.file_operations as fo
import uuid

from ResSimpy.Utils.factory_methods import get_empty_list_file


@dataclass(kw_only=True)
class File(FileBase):
    """The abstract base class for simulator files.

    Attributes:
        location (str): Full path to file location
        file_content_as_list (list[str]): List of lines in the file
    """

    location: str
    _location_in_including_file: str
    include_objects: Optional[Sequence[File]]
    include_locations: Optional[list[str]] = None
    file_content_as_list: Optional[list[str]] = field(default=None, repr=False)
    __id: UUID = field(default_factory=lambda: uuid4(), compare=False)
    __file_modified: bool = False
    __file_loading_skipped: bool = False

    def __init__(self, location: str,
                 file_content_as_list: Optional[list[str]] = None,
                 include_objects: Optional[Sequence[File]] = None, create_as_modified: bool = False,
                 file_loading_skipped: bool = False) -> None:
        """Initialises the File class.

        Args:
            location (str): The location of the file on disk.
            file_content_as_list (Optional[list[str]]): The file content as a list of strings.
            include_objects (Optional[Sequence[File]]): The files included in the file.
            create_as_modified (bool): Set the object to modified.
            file_loading_skipped (bool): Whether the file loading was skipped due to the file being too large an array.
        """
        self.location = location
        self._location_in_including_file = location
        self.include_objects: Optional[list[File]] = get_empty_list_file() \
            if include_objects is None else include_objects
        if file_content_as_list is None:
            self.file_content_as_list = []
        else:
            self.file_content_as_list = file_content_as_list
        self.__id = uuid.uuid4()
        self.__file_modified = create_as_modified
        self.__file_loading_skipped = file_loading_skipped

    @property
    def id(self) -> UUID:
        """Unique identifier for each Node object."""
        return self.__id

    @property
    def file_modified(self) -> bool:
        """Whether the file has been modified since the last file write."""
        return self.__file_modified

    @property
    def file_loading_skipped(self) -> bool:
        """Whether loading the file contents has been skipped due to it being to large e.g. for array files."""
        return self.__file_loading_skipped

    @property
    def location_in_including_file(self) -> str:
        """The location of the file as it is written after the INCLUDE token in the file including it."""
        return self._location_in_including_file

    def update_include_location_in_file_as_list(self, new_path: str, include_file: File) -> None:
        """update_include_location_in_file_as_list.

        Args:
            new_path(str): Updates the path in the file as list to point towards the new include location.
            include_file(file): include object whose path is being modified.
        """
        raise NotImplementedError("Implement this in the derived class.")

    def write_to_file(self, new_file_path: None | str = None, write_includes: bool = False,
                      write_out_all_files: bool = False, overwrite_file: bool = False) -> None:
        """Writes to file specified in self.location the strings contained in the list self.file_content_as_list.

        Args:
            new_file_path (None | str): writes to self.location if left as None. Otherwise writes to new_file_name.
            write_includes (bool): If True will write out all include files within the file. Defaults to False.
            write_out_all_files (bool): If False will write only modified files. Otherwise will write all files.
            overwrite_file (bool): If True will overwrite the file at the location specified by new_file_path. \
            Otherwise will raise an error if the file already exists.
        """
        # overwrite File base class method to allow for write_includes
        if new_file_path is None:
            if overwrite_file:
                # In this case just overwrite the file with the existing path:
                new_file_path = self.location
            else:
                raise ValueError('No file path to write to, and overwrite_file set to False')

        if self.file_content_as_list is None:
            raise ValueError(f'No file data to write out, instead found {self.file_content_as_list}')

        # create directories that do not exist
        if not os.path.exists(os.path.dirname(new_file_path)):
            os.makedirs(os.path.dirname(new_file_path))

        if write_includes and self.include_objects is not None:
            for file in self.include_objects:
                write_file: bool = file.file_modified or write_out_all_files
                write_file = write_file or (new_file_path != self.location and not os.path.isabs(file.location))

                # if the array was previously skipped then load the file as list
                if file.file_loading_skipped:
                    file.file_content_as_list = fo.load_file_as_list(file.location)

                if file.file_content_as_list is None:
                    warnings.warn(f'No content found for file: {file.location}. Not writing file.')
                    continue
                # this distinguishes the include file name from other includes in other files.
                # this could be moved to a folder name instead rather than affecting the include file name?
                new_root_name = f'{os.path.basename(new_file_path).split(".")[0]}_{os.path.basename(file.location)}'
                # write the include file to the same directory.
                # TODO add a fix for the name above and a check for whether the file exists as exactly the same name
                include_file_name = os.path.join(os.path.dirname(new_file_path), new_root_name)

                if write_file:
                    self.update_include_location_in_file_as_list(include_file_name, file)

                if os.path.exists(include_file_name):
                    # if the file is already copied across then move on to the next file
                    continue

                if write_file:
                    file.write_to_file(include_file_name, write_includes=True, write_out_all_files=write_out_all_files,
                                       overwrite_file=overwrite_file)

        # check if the file already exists so we aren't accidentally overwriting it
        if os.path.exists(new_file_path) and not overwrite_file:
            raise ValueError(f'File already exists at {new_file_path} and overwrite_file set to False')

        file_str = ''.join(self.file_content_as_list)

        # update the location:
        if self.file_modified or write_out_all_files:
            self.location = new_file_path

        # write the file to the new location
        with open(new_file_path, 'w') as fi:
            fi.write(file_str)
        # reset the modified file state
        self._file_modified_set(False)

    def _file_modified_set(self, value: bool) -> None:
        """Set the modified file status.

        Args:
            value(bool): True if the file is modified, False otherwise.
        """
        self.__file_modified = value

    @property
    def get_flat_list_str_file(self) -> list[str]:
        """Returns flat list of strings from file."""
        raise NotImplementedError("Implement this in the derived class")

    def add_object_locations(self, obj_uuid: UUID, line_indices: list[int]) -> None:
        """Adds a uuid to the object_locations dictionary.

        Args:
            obj_uuid(UUID): Unique identifier of the object being created/stored.
            line_indices(list[int]): line number in the flattened file_content_as_list
            (i.e. from the get_flat_list_str_file method).
        """
        raise NotImplementedError("Implement this in the derived class")

    @staticmethod
    def insert_comments(additional_content: list[str], comments: str) -> list[str]:
        """Adds comments alongside additional content.

        Args:
            additional_content(list[str]): additional lines of the file to be added with a new entry per line.
            comments(str): comments to be added to all lines list of strings within the content.
        """
        raise NotImplementedError("Implement this in the derived class")

    def get_object_locations_for_id(self, obj_id: UUID) -> list[int]:
        """Returns the list of line numbers associated with the object ID.

        Args:
            obj_id(UUID): Unique identifier for each node object.
        """
        raise NotImplementedError("Implement this in the derived class")

    def remove_object_from_file_as_list(self, objects_to_remove: list[UUID]) -> None:
        """Removes all associated lines in the file as well as the object locations relating to a list of objects.

        Args:
            objects_to_remove(list[UUID]):  list of object ids to remove from the object locations. Defaults to None
        """
        raise NotImplementedError("Implement this in the derived class")

    def add_to_file_as_list(self, additional_content: list[str], index: int,
                            additional_objects: Optional[dict[UUID, list[int]]] = None,
                            comments: Optional[str] = None) -> None:
        """To add content to the file as list, also updates object numbers
        and optionally allows user to add several additional new objects.

        Args:
            additional_content(list[str]):  Additional lines as a list of strings to be added.
            index(int): index to insert the new lines at in the calling flat_file_as_list
            additional_objects(Optional[dict[UUID, list[int]]]): defaults to None. Otherwise,
            a dictionary keyed with the UUID of the new objects to add as well as
            the corresponding index of the object in the original calling NexusFile.
            comments(Optional[str]): defaults to None. Comments to add in-line to the file.
        """
        raise NotImplementedError("Implement this in the derived class")

    def remove_from_file_as_list(self, index: int, objects_to_remove: Optional[list[UUID]] = None,
                                 string_to_remove: Optional[str] = None) -> None:
        """Remove an entry from the file as list.

        Args:
            index(int): index n the calling flat_file_as_list to remove the entry from.
            objects_to_remove(Optional[list[UUID]): list of object ids to
            remove from the object locations. Defaults to None.
            string_to_remove(Optional[str]):if specified will only remove the listed string from the entry at the index.
             Defaults to None, which removes the entire entry.
        """
        raise NotImplementedError("Implement this in the derived class")

    def find_which_include_file(self, flattened_index: int) -> tuple[File, int]:
        """Given a line index that relates to a position within the flattened
        file_as_list from the method get_flat_file_as_list.

        Args:
            flattened_index(int): index in the flattened file as list structure.
        """
        raise NotImplementedError("Implement this in the derived class")

    def __repr__(self) -> str:
        """A more readable representation of the NexusFile object file data."""
        repr_string = ''

        repr_string += f'FILE PATH: {self.location}\n\n'
        repr_string += f'Include files: {self.include_locations}\n\n'

        repr_string += 'FILE CONTENTS:\n\n'
        repr_string += self.pretty_print_contents()
        return repr_string

    def pretty_print_contents(self) -> str:
        """Pretty print the file contents."""
        if self.file_content_as_list is None:
            return ''
        return ''.join(self.file_content_as_list)



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Any, Sequence, Optional, TYPE_CHECKING, TypeVar
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixin
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import Network

T = TypeVar('T', bound=DataObjectMixin)


class NetworkOperationsMixIn(ABC):
    """A mixin for network collections that forces the implementation of the same methods across all classes."""
    def __init__(self, parent_network: Network) -> None:
        """Initialises the NetworkOperationsMixIn class.

        Args:
            parent_network (Network): The parent network that the object is a part of.
        """
        self.__parent_network = parent_network

    @abstractmethod
    def get_all(self) -> Sequence[Any]:
        """Returns list of network operations."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_by_name(self, name: str) -> Optional[Any]:
        """Returns network operation with provided name.

        Args:
            name(str): Name of network operation.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_df(self) -> pd.DataFrame:
        """Creates dataframe representing network operations in a file."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_overview(self) -> str:
        """Returns an overview of the network operations."""
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load(self, file: File, start_date: str, default_units: UnitSystem) -> None:
        """Loads file path using the default units.

        Args:
            file(File): Loads file path
            start_date(str): Starting date of the run.
            default_units(UnitSystem): Units used if not specified by the file.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def _add_to_memory(self, additional_objs: Optional[list[Any]]) -> None:
        """Adds an instance of objects to the current list.

        Args:
            additional_objs(Optional[list[Any]]): New objects to be added.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove(self, obj_to_remove: UUID | dict[str, None | str | float | int]) -> None:
        """Removes the UUID from the network.

        Args:
            obj_to_remove(UUID | dict[str, None | str | float | int]): UUID of the network to be removed
            from the dictionary.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add(self, obj_to_add: dict[str, None | str | float | int]) -> None:
        """Adding a connection to a network taking a dictionary with properties for the new connection.

        Args:
            obj_to_add(dict[str, None | str | float | int]): dictionary taking all the new properties for the network.
        """
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def modify(self, obj_to_modify: dict[str, None | str | float | int],
               new_properties: dict[str, None | str | float | int]) -> None:
        """Modify existing object based on a matching dictionary of properties.

        Args:
            obj_to_modify(dict[str, None | str | float | int]): dictionary containing attributes to match
            in the existing object.
            new_properties(dict[str, None | str | float | int]): properties to switch to in the new object.
        """
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_header(self) -> str:
        """Returns table header as a string."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_footer(self) -> str:
        """Returns table footer as a string."""
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def _network_element_name(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @staticmethod
    def resolve_carried_over_attributes(objects_to_resolve: Sequence[T]) -> Sequence[T]:
        """Resolves carried over attributes from previous time steps to the future timesteps.

        Args:
            objects_to_resolve (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.
        """
        resolved_objects: list[T] = []

        sorted_by_date, unique_names = NetworkOperationsMixIn.__sort_by_date_name_sim_order(objects_to_resolve)

        # resolve by name
        for name in unique_names:
            matching_names = [x for x in sorted_by_date if x.name == name]
            resolving_by_name = NetworkOperationsMixIn.resolve_same_named_objects(matching_names)
            resolved_objects.extend(resolving_by_name)

        return resolved_objects

    @staticmethod
    def __sort_by_date_name_sim_order(objects_to_resolve: Sequence[T]) -> tuple[Sequence[T], list[str | None]]:
        """Sorts objects by date and name.

        Args:
            objects_to_resolve (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.
        """
        if (len(objects_to_resolve) > 0 and
                not all(isinstance(x, type(objects_to_resolve[0])) for x in objects_to_resolve)):
            raise ValueError("Objects to resolve must be of the same type.")
        # order by date and the order entered in the simulator.
        current_ordering = list(enumerate(objects_to_resolve))
        sorted_by_date_sim_ordering = sorted(current_ordering, key=lambda x: (x[1].iso_date, x[0]))
        sorted_by_date = [x[1] for x in sorted_by_date_sim_ordering]
        # split by the name of the object
        unique_names = list({x.name for x in sorted_by_date})
        return sorted_by_date, unique_names

    @staticmethod
    def resolve_same_named_objects(sorted_by_date: Sequence[T]) -> Sequence[T]:
        """Resolves a subset of objects by date.

        Args:
            sorted_by_date (Sequence[DataObjectMixin]): list of objects to resolve carried over attributes for.
            Must be of homogenous type.

        Returns:
            Sequence[DataObjectMixin]: list of resolved objects with carried over attributes.

        """
        resolved_objects: list[T] = []
        for unresolved_obj in sorted_by_date:
            # append the first
            if len(resolved_objects) == 0:
                resolved_objects.append(unresolved_obj)
                continue
            last_resolved_object = resolved_objects[-1]

            new_resolved_object = copy.deepcopy(unresolved_obj)

            for attr, value in last_resolved_object.__dict__.items():
                if value is not None and getattr(unresolved_obj, attr, None) is None:
                    setattr(new_resolved_object, attr, value)
            resolved_objects.append(new_resolved_object)
        return resolved_objects

    @staticmethod
    def resolve_same_named_objects_constraints(sorted_by_data: Sequence[NexusConstraint]) -> Sequence[NexusConstraint]:
        """Resolves a subset of objects by date and applies clears in the constraints.

        Args:
            sorted_by_data (Sequence[NexusConstraint]): Sequence of constraints to resolve in .
        """
        resolved_objects: list[NexusConstraint] = []
        for unresolved_obj in sorted_by_data:
            # append the first
            if len(resolved_objects) == 0:
                resolved_objects.append(unresolved_obj)
                continue
            last_resolved_object = resolved_objects[-1]
            # copy to prevent overriding previously resolved objects
            last_resolved_copy = copy.deepcopy(last_resolved_object)
            new_resolved_object = copy.deepcopy(unresolved_obj)

            # collect which attributes to clear:
            clear_constraints_dict = {}
            if unresolved_obj.clear_q:
                clear_constraints_dict.update(unresolved_obj.get_rate_constraints_map())
            if unresolved_obj.clear_p:
                clear_constraints_dict.update(unresolved_obj.get_pressure_constraints_map())
            if unresolved_obj.clear_limit:
                clear_constraints_dict.update(unresolved_obj.get_limit_constraints_map())
            if unresolved_obj.clear_alq:
                clear_constraints_dict.update(unresolved_obj.get_alq_constraints_map())
            if unresolved_obj.clear_all:
                clear_constraints_dict.update(unresolved_obj.get_rate_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_pressure_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_limit_constraints_map())
                clear_constraints_dict.update(unresolved_obj.get_alq_constraints_map())
            # Clear them by setting to None on the copied previously resolved object
            for (clear_attr, _) in clear_constraints_dict.values():
                setattr(last_resolved_copy, clear_attr, None)

            skip_attributes = ['id', 'date', 'name', 'iso_date', 'clear_q', 'clear_p', 'clear_limit', 'clear_alq',
                               'clear_all', 'convert_qmult_to_reservoir_barrels']
            for attr, value in last_resolved_copy.__dict__.items():
                if value is None or attr in skip_attributes:
                    continue
                if getattr(unresolved_obj, attr, None) is None:
                    setattr(new_resolved_object, attr, value)

            # resolve convert_qmult_to_reservoir_barrels
            if last_resolved_object.convert_qmult_to_reservoir_barrels:
                overriding_qmult_constraints = [unresolved_obj.max_surface_gas_rate,
                                                unresolved_obj.max_surface_water_rate,
                                                unresolved_obj.max_surface_oil_rate,
                                                ]
                if any(x for x in overriding_qmult_constraints if x is not None):
                    # clear any qmult constraints if any of the surface rate constraints for oil, gas, water are set
                    new_resolved_object.convert_qmult_to_reservoir_barrels = None
                    new_resolved_object.qmult_oil_rate = None
                    new_resolved_object.qmult_gas_rate = None
                    new_resolved_object.qmult_water_rate = None
                else:
                    new_resolved_object.convert_qmult_to_reservoir_barrels = True

            resolved_objects.append(new_resolved_object)
        return resolved_objects



"""Functions for collecting tables from a Nexus file to populate Nexus objects."""
from __future__ import annotations

import fnmatch
from datetime import timedelta, time
from typing import Any, Optional

import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.FileOperations.File import File
from ResSimpy.FileOperations.file_operations import get_next_value
from ResSimpy.Nexus.DataModels.Network.NexusActivationChange import NexusActivationChange
from ResSimpy.Nexus.DataModels.Network.NexusConLists import NexusConLists
from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Nexus.DataModels.Network.NexusWellLists import NexusWellLists
from ResSimpy.Nexus.DataModels.NexusWellList import NexusWellList
from ResSimpy.Nexus.NexusEnums.ActivationChangeEnum import ActivationChangeEnum
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.nexus_constraint_operations import load_inline_constraints
from ResSimpy.Nexus.nexus_file_operations import check_property_in_line, check_token, get_expected_token_value, \
    check_list_tokens, load_table_to_objects
from ResSimpy.Nexus.nexus_load_list_table import load_table_to_lists
from ResSimpy.Time.ISODateTime import ISODateTime


# TODO refactor the collection of tables to an object with proper typing
def collect_all_tables_to_objects(nexus_file: File, table_object_map: dict[str, Any], start_date: Optional[str],
                                  default_units: Optional[UnitSystem], date_format: DateFormat) -> \
        tuple[dict[str, list[Any]], dict[str, list[NexusConstraint]]]:
    """Loads all tables from a given file.

    Args:
    ----
        nexus_file (File): NexusFile representation of the file.
        table_object_map (dict[str, Storage_Object]): dictionary containing the name of the table as keys and \
                    the object type to store the data from each row into. Require objects to have a get_keyword_mapping\
                    function
        model: (NexusSimulator): main simulator object
        start_date (Optional[str]): The model start date.
        default_units (Optional[UnitSystem]): The default unit system the Nexus model is set to.
        date_format (Optional[DateFormat]): The date format of the object.

    Raises:
    ------
    TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

    Returns:
    -------
    tuple[dict[str, list[DataObject]], dict[str, list[NexusConstraint]]]: a tuple of two dictionaries of lists of \
    DataObjects. The first element is a dictionary of lists of objects keyed with the NexusTable name associated with \
    the object. The second element is a dictionary of lists of NexusConstraints keyed with the well name associated \
    with the constraint.
    """
    current_date = start_date
    nexus_object_results: dict[str, list[Any]] = {x: [] for x in table_object_map}
    nexus_constraints: dict[str, list[NexusConstraint]] = {}

    file_as_list: list[str] = nexus_file.get_flat_list_str_file
    table_start: int = -1
    table_end: int = -1
    property_dict: dict = {}
    token_found: Optional[str] = None
    network_names: list[str] = []
    well_names: list[str] = []
    well_lists: list[NexusWellList] = []
    is_deactivate_block = False
    is_activate_block = False
    is_actions_block = False
    is_constraints_block = False
    for index, line in enumerate(file_as_list):
        # check for changes in unit system
        check_property_in_line(line, property_dict, file_as_list)
        unit_system = property_dict.get('UNIT_SYSTEM', default_units)
        if not isinstance(unit_system, UnitSystem):
            raise TypeError(f"Value found for {unit_system=} of type {type(unit_system)} \
                                not compatible, expected type UnitSystem Enum")

        if current_date is None:
            raise ValueError("Unable to determine date for object")

        # Important!: DEACTIVATE can also appear in the ACTIONS table in the surface network and CONSTRAINTS tables.
        # If we are already in one of those blocks, we don't want to consider it a new activate / deactivate block.
        if fo.check_token('ACTIONS', line=line):
            is_actions_block = True
        if fo.check_token('ENDACTIONS', line=line):
            is_actions_block = False
        if fo.check_token('CONSTRAINTS', line=line):
            is_constraints_block = True
        if fo.check_token('ENDCONSTRAINTS', line=line):
            is_constraints_block = False

        if not is_actions_block and not is_constraints_block:
            is_activate_block, is_deactivate_block, should_continue = \
                __activate_deactivate_checks(line=line,
                                             is_activate_block=is_activate_block,
                                             is_deactivate_block=is_deactivate_block,
                                             current_date=current_date,
                                             start_date=start_date,
                                             date_format=date_format,
                                             nexus_object_results=nexus_object_results)

            if should_continue:
                # If we have already read in all the values for this line in the previous call, continue onto the next
                # line.
                continue

        if check_token('TIME', line) and table_start < 0:
            time_value = get_expected_token_value(
                token='TIME', token_line=line, file_list=file_as_list,
                custom_message=f"Cannot find the date associated with the TIME card in {line=} at line number {index}")
            if time_value.upper() != 'PLUS':
                current_date = time_value
            else:
                plus_value = get_expected_token_value(token='PLUS', token_line=line, file_list=file_as_list,
                                                      custom_message="Cannot find the date associated with the TIME "
                                                                     f"PLUS card in {line=} at line number {index}")

                if current_date is None:
                    raise ValueError("Cannot calculate PLUS date without access to the initial date.")
                else:
                    new_datetime = ISODateTime.convert_to_iso(
                        current_date, date_format, start_date) + timedelta(days=float(plus_value))

                if date_format == DateFormat.DD_MM_YYYY:
                    if new_datetime.time() == time(0, 0, 0, 0):
                        current_date = new_datetime.strftime('%d/%m/%Y')
                    else:
                        current_date = new_datetime.strftime('%d/%m/%Y(%H:%M:%S)')
                elif date_format == DateFormat.MM_DD_YYYY:
                    if new_datetime.time() == time(0, 0, 0, 0):
                        current_date = new_datetime.strftime('%m/%d/%Y')
                    else:
                        current_date = new_datetime.strftime('%m/%d/%Y(%H:%M:%S)')
            continue
        if table_start < 0:
            token_found = check_list_tokens(list(table_object_map.keys()), line)
            if token_found is None or check_token('WELLCONTROL', line):
                continue
            # if a token is found get the starting index of the table
            table_start = index + 1
        if token_found is None:
            continue
        token_found = token_found.upper()

        if table_start > 0 and check_token('END' + token_found, line):
            table_end = index
        # if we have a complete table to read in start reading it into objects
        if 0 < table_start <= table_end:
            # cover for the case where we aren't currently reading in the table to an object.
            # if no object is provided by the map for the token found then skip the keyword and reset the tracking vars
            if table_object_map[token_found] is None:
                table_start = -1
                table_end = -1
                token_found = None
                continue
            list_objects = None
            property_map = table_object_map[token_found].get_keyword_mapping()
            if token_found == 'CONSTRAINTS':
                load_inline_constraints(file_as_list=file_as_list[table_start:table_end],
                                        constraint=table_object_map[token_found],
                                        current_date=current_date,
                                        unit_system=unit_system,
                                        property_map=property_map,
                                        existing_constraints=nexus_constraints,
                                        nexus_file=nexus_file,
                                        start_line_index=table_start,
                                        network_names=network_names,
                                        date_format=date_format,
                                        welllists=well_lists,
                                        start_date=start_date)

            elif token_found == 'QMULT' or token_found == 'CONSTRAINT':
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system,
                                                     constraint_obj_dict=nexus_constraints,
                                                     preserve_previous_object_attributes=True,
                                                     date_format=date_format,
                                                     welllists=well_lists,
                                                     start_date=start_date)

            elif token_found in [NexusWellLists.table_header(), NexusConLists.table_header()]:
                list_objects = load_table_to_lists(file_as_list=file_as_list[table_start - 1:table_end],
                                                   row_object=table_object_map[token_found],
                                                   table_header=token_found,
                                                   current_date=current_date,
                                                   previous_lists=nexus_object_results[token_found],
                                                   date_format=date_format)

                # If there is already a matching Welllist for this timestamp, update that with the additional changes,
                # rather than adding another one.
                existing_welllist = next(
                    (x for x in nexus_object_results[NexusWellLists.table_header()] if
                     x.name == list_objects[0][0].name and
                     x.date == list_objects[0][0].date), None)

                if existing_welllist is not None:
                    existing_welllist.elements_in_the_list = list_objects[0][0].wells
                    list_objects = []
                if token_found == NexusWellLists.table_header():
                    well_lists = [x[0] for x in list_objects]

            else:
                list_objects = load_table_to_objects(file_as_list=file_as_list[table_start:table_end],
                                                     row_object=table_object_map[token_found],
                                                     property_map=property_map,
                                                     current_date=current_date,
                                                     unit_system=unit_system, date_format=date_format,
                                                     well_names=well_names, start_date=start_date)

            # store objects found into right dictionary
            list_of_token_obj = nexus_object_results[token_found]
            # This statement ensures that CONSTRAINT that are found in tables are actually added to the dictionary
            # under the same key as constraints to preserve their order
            if (token_found == 'CONSTRAINT' or token_found == 'QMULT') and list_objects is not None:
                for constraint, id_index in list_objects:
                    if isinstance(constraint, NexusConstraint):
                        obj_id = constraint.id
                        well_name = constraint.name
                    else:
                        obj_id = constraint
                        well_name = None
                    correct_line_index = id_index + table_start
                    nexus_file.add_object_locations(obj_id, [correct_line_index])
                    if well_name is None:
                        continue
                    if nexus_constraints.get(well_name, None) is not None:
                        nexus_constraints[well_name].append(constraint)
                    else:
                        nexus_constraints[well_name] = [constraint]

            elif list_objects is not None and isinstance(list_of_token_obj, list):
                list_of_token_obj.extend([x[0] for x in list_objects])
                # add the names from the nodes into the network names for wildcards
                # wildcards do not apply to LIST type nexus objects like WELLLIST.
                if 'LIST' not in token_found:
                    network_names.extend([x.name for x in list_of_token_obj])
                if token_found == 'WELLS':
                    well_names.extend([x.name for x in list_of_token_obj if x.name not in well_names])
                for new_object, id_index in list_objects:
                    correct_line_index = id_index + table_start
                    # temporary try statement until all objects have an id property
                    try:
                        obj_id = new_object.id
                        nexus_file.add_object_locations(obj_id, [correct_line_index])
                    except AttributeError:
                        pass
            # reset indices for further tables
            table_start = -1
            table_end = -1
            token_found = None
    return nexus_object_results, nexus_constraints


def __activate_deactivate_checks(line: str, is_activate_block: bool,
                                 is_deactivate_block: bool, current_date: str, start_date: str | None,
                                 date_format: DateFormat, nexus_object_results: dict[str, list[Any]]) \
        -> tuple[bool, bool, bool]:
    """Checks if the line being read is the start of an activate / deactivate block, and if it is, sets the property on
    the related Wellconnection objects found.

    Returns:
    tuple[bool, bool, bool]: a tuple of three bools, stating whether the following text is an activate block, whether \
        it is a deactivate block, and whether the loop calling this method should continue to the next line.
    """
    # Handle activation / deactivation of well connections
    if check_token('DEACTIVATE', line):
        is_deactivate_block = True
        return is_activate_block, is_deactivate_block, True

    if check_token('ACTIVATE', line):
        is_activate_block = True
        return is_activate_block, is_deactivate_block, True

    if not is_activate_block and not is_deactivate_block:
        return is_activate_block, is_deactivate_block, False

    if check_token('CONNECTION', line):
        return is_activate_block, is_deactivate_block, True

    if check_token('ENDDEACTIVATE', line) or check_token('ENDACTIVATE', line):
        is_deactivate_block = False
        is_activate_block = False
        return is_activate_block, is_deactivate_block, True

    # loop through all the connection names found in the block, and create the ActivationChange objects.
    found_connections = []
    affected_connection_name = get_next_value(start_line_index=0, file_as_list=[line])

    # Get a list of all the distinct connection names that might be affected.
    all_well_connection_names = []
    connection_types = {'WELLS', 'GASWELLS', 'NODECON'} & set(nexus_object_results)
    for connection_type in connection_types:
        all_well_connection_names.extend([connection.name for
                                          connection in nexus_object_results[connection_type]])
    all_well_connections = list(dict.fromkeys(all_well_connection_names))

    while affected_connection_name is not None:
        change = ActivationChangeEnum.ACTIVATE if is_activate_block else ActivationChangeEnum.DEACTIVATE
        # If a wildcard is found, ensure that all connections that match the wildcard are dealt with.
        if '*' in affected_connection_name:
            all_affected_well_connections = [x for x in all_well_connections if
                                             fnmatch.fnmatch(x.casefold(), affected_connection_name.casefold())]

            # Create an object for each affected connection
            for connection in all_affected_well_connections:
                new_activation_change = NexusActivationChange(change=change, name=connection,
                                                              date=current_date,
                                                              start_date=start_date, date_format=date_format)
                nexus_object_results['ACTIVATE_DEACTIVATE'].append(new_activation_change)

        else:
            new_activation_change = NexusActivationChange(change=change, name=affected_connection_name,
                                                          date=current_date, start_date=start_date,
                                                          date_format=date_format)
            nexus_object_results['ACTIVATE_DEACTIVATE'].append(new_activation_change)

        found_connections.append(affected_connection_name)
        # Move to the next value
        affected_connection_name = get_next_value(start_line_index=0, file_as_list=[line],
                                                  ignore_values=found_connections)

    return is_activate_block, is_deactivate_block, False



"""Abstract base class for aquifer inputs."""
from dataclasses import dataclass
from abc import ABC
from typing import Mapping

from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty


@dataclass(kw_only=True)
class Aquifer(ABC):
    """The abstract base class for a collection of aquifer inputs.

    Attributes:
        inputs (dict[int, DynamicProperty]): Collection of aquifer inputs, as a dictionary.
    """

    @property
    def inputs(self) -> Mapping[int, DynamicProperty]:
        """A Collection of aquifer inputs, as a dictionary."""
        raise NotImplementedError("Implement this in the derived class")



"""Nexus implementation of the Well Connection class."""

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.DataModelBaseClasses.WellConnection import WellConnection


@dataclass(kw_only=True)
class NexusWellConnection(WellConnection):
    """Nexus implementation of the Well Connection class.

    Attributes:
    stream (str): Stream identifier (STREAM)
    number (int): Identification number (NUMBER)
    scale (float): Scaling factor (SCALE)
    bhdepth (float): Bottomhole depth (BHDEPTH)
    datum_depth (float): Depth relative to a datum (DATUM)
    x_pos (float): X-coordinate position (X)
    y_pos (float): Y-coordinate position (Y)
    gradient_calc (str): Gradient calculation method (DATGRAD)
    length (float): Length of the well connection (LENGTH)
    bottomhole_measured_depth (float): Measured depth of the bottom hole (BHMD)
    add_tubing (int): Additional tubing indicator (ADDTUBING)
    diameter (float): Diameter of the well connection (DIAM)
    inner_diameter (float): Inner diameter of the well connection (INNERDIAM)
    roughness (float): Pipe roughness (ROUGHNESS)
    tracer (str): Tracer substance identifier (TRACERS)
    con_type (str): Connection type (TYPE)
    hyd_method (str): Hydraulic method (METHOD)
    pvt_method (int): PVT method (IPVT)
    water_method (int): Water method (IWAT)
    bat_method (int): Bat method (IBAT)
    temperature (float): Temperature (TEMP)
    elevation_profile (str): Elevation profile identifier (ELEVPR)
    temperature_profile (str): Temperature profile identifier (TEMPPR)
    inj_mobility (str): Injection mobility identifier (INJMOB)
    crossshut (str): Cross-shut identifier (CROSS_SHUT)
    crossflow (str): Crossflow identifier (CROSSFLOW)
    on_time (float): On-time duration (ONTIME)
    heat_transfer_coeff (float): Heat transfer coefficient (HTC)
    well_index_mult (float): Well index multiplier (WIMULT)
    productivity_index (float): Productivity index (PI)
    vip_productivity_index (str): VIP productivity index identifier (VIPPI)
    productivity_index_phase (str): Productivity index phase identifier (PIPHASE)
    d_factor (float): Non-Darcy D-factor (D)
    non_darcy_flow_model (str): Non-Darcy flow model identifier (ND)
    non_darcy_flow_method (str): Non-Darcy flow method identifier (DPERF)
    gas_mobility (float): Gas mobility value (GASMOB)
    capillary_number_model (str): Capillary number model identifier (CN)
    dp_add (float): Additional pressure drop (DPADD)
    dt_add (float): Additional temperature change (DTADD)
    rate_mult (float): Rate multiplier (RATEMULT)
    polymer (str): Polymer identifier (POLYMER)
    station (str): Station identifier (STATION)
    drill_queue (str): Drill queue identifier (ASSCDR)
    drill_order_benefit (float): Benefit of the drill order (BENEFIT)
    is_activated (bool): Whether the Well Connection has been activated using ACTIVATED / DEACTIVATE. Defaults to True.
    """

    bh_node_name: Optional[str] = None
    wh_node_name: Optional[str] = None

    stream: Optional[str] = None
    number: Optional[int] = None
    scale: Optional[float] = None
    gradient_calc: Optional[str] = None
    bottomhole_measured_depth: Optional[float] = None
    add_tubing: Optional[int] = None
    tracer: Optional[str] = None
    con_type: Optional[str] = None
    pvt_method: Optional[int] = None
    water_method: Optional[int] = None
    bat_method: Optional[int] = None
    elevation_profile: Optional[str] = None
    temperature_profile: Optional[str] = None
    inj_mobility: Optional[str] = None
    crossshut: Optional[str] = None
    crossflow: Optional[str] = None
    on_time: Optional[float] = None
    heat_transfer_coeff: Optional[float] = None
    well_index_mult: Optional[float] = None
    vip_productivity_index: Optional[float] = None
    productivity_index_phase: Optional[str] = None
    d_factor: Optional[float] = None
    non_darcy_flow_model: Optional[str] = None
    non_darcy_flow_method: Optional[str] = None
    gas_mobility: Optional[float] = None
    capillary_number_model: Optional[str] = None
    dp_add: Optional[float] = None
    dt_add: Optional[float] = None
    rate_mult: Optional[float] = None
    polymer: Optional[str] = None
    station: Optional[str] = None
    drill_queue: Optional[str] = None
    drill_order_benefit: Optional[float] = None

    def __init__(self, properties_dict: Optional[dict[str, None | int | str | float]] = None,
                 date: Optional[str] = None, date_format: Optional[DateFormat] = None, start_date: Optional[str] = None,
                 unit_system: Optional[UnitSystem] = None, name: Optional[str] = None,
                 bh_node_name: Optional[str] = None, wh_node_name: Optional[str] = None,
                 stream: Optional[str] = None, number: Optional[int] = None, scale: Optional[float] = None,
                 gradient_cal: Optional[str] = None, bottomhole_measured_depth: Optional[float] = None,
                 add_tubing: Optional[int] = None, tracer: Optional[str] = None, con_type: Optional[str] = None,
                 pvt_method: Optional[int] = None, water_method: Optional[int] = None, bat_method: Optional[int] = None,
                 elevation_profile: Optional[str] = None, temperature_profile: Optional[str] = None,
                 inj_mobility: Optional[str] = None, crossshut: Optional[str] = None, crossflow: Optional[str] = None,
                 on_time: Optional[float] = None, heat_transfer_coeff: Optional[float] = None,
                 well_index_mult: Optional[float] = None, vip_productivity_index: Optional[float] = None,
                 productivity_index_phase: Optional[str] = None, d_factor: Optional[float] = None,
                 non_darcy_flow_model: Optional[str] = None, non_darcy_flow_method: Optional[str] = None,
                 gas_mobility: Optional[float] = None, capillary_number_model: Optional[str] = None,
                 dp_add: Optional[float] = None, dt_add: Optional[float] = None, rate_mult: Optional[float] = None,
                 polymer: Optional[str] = None, station: Optional[str] = None, drill_queue: Optional[str] = None,
                 drill_order_benefit: Optional[float] = None, bhdepth: Optional[float] = None,
                 datum_depth: Optional[float] = None, x_pos: Optional[float] = None,
                 y_pos: Optional[float] = None, length: Optional[float] = None, temperature: Optional[float] = None,
                 diameter: Optional[float] = None, roughness: Optional[float] = None,
                 inner_diameter: Optional[float] = None, productivity_index: Optional[float] = None,
                 hyd_method: Optional[str] = None, group: Optional[str] = None, i: Optional[int] = None,
                 j: Optional[int] = None, drainage_radius: Optional[float] = None) -> None:
        """Initialises the NexusWellConnection class.

        Args:
            properties_dict (Optional[dict]): dict of the properties to set on the object.
            date (Optional[str]): The date of the object.
            date_format (Optional[DateFormat]): The date format that the object uses.
            start_date (Optional[str]): The start date of the model. Required if the object uses a decimal TIME.
            unit_system (Optional[UnitSystem]): The unit system of the object e.g. ENGLISH, METRIC.
            is_activated: bool: Whether the Well Connection has been set as activated. Defaults to True.
            name (Optional[str]): The name of the object.
            bh_node_name (Optional[str]): Bottom hole node name.
            wh_node_name (Optional[str]): Wellhead node name.
            stream (Optional[str]): Stream identifier (STREAM).
            number (Optional[int]): Identificaiton number (Number).
            scale (Optional[float]): Scaling factor (SCALE).
            gradient_cal (Optional[str]): Gradient calculation method (DATGRAD).
            bottomhole_measured_depth (Optional[float]): Measured depth of the bottom hole (BHMD).
            add_tubing (Optional[int]): Additional tubing indicator (ADDTUBING).
            tracer (Optional[str]): Tracer substance identifier (TRACERS).
            con_type (Optional[str]): Connection type (TYPE).
            pvt_method (Optional[int]): PVT method (IPVT).
            water_method (Optional[int]): Water method (IWAT).
            bat_method (Optional[int]): Bat method (IBAT).
            elevation_profile (Optional[str]): Elevation profile identifier (ELEVPR).
            temperature_profile (Optional[str]): Temperature profile identifier (TEMPPR).
            inj_mobility (Optional[str]): Injection mobility identifier (INJMOB).
            crossshut (Optional[str]): Cross-shut identifier (CROSS_SHUT).
            crossflow (Optional[str]): Crossflow identifier (CROSSFLOW).
            on_time (Optional[float]): On-time duration (ONTIME).
            heat_transfer_coeff (Optional[float]): Heat transfer coefficient (HTC).
            well_index_mult (Optional[float]): Well index multiplier (WIMULT).
            vip_productivity_index (Optional[float]): VIP productivity index identifier (VIPPI).
            productivity_index_phase (Optional[str]): Productivity index phase identifier (PIPHASE).
            d_factor (Optional[float]): Non-Darcy D-factor (D).
            non_darcy_flow_model (Optional[str]): Non-Darcy flow model identifier (ND).
            non_darcy_flow_method (Optional[str]): Non-Darcy flow method identifier (DPERF)
            gas_mobility (Optional[float]): Gas mobility value (GASMOB).
            capillary_number_model (Optional[str]): Capillary number model identifier (CN).
            dp_add (Optional[float]): Additional pressure drop (DPADD).
            dt_add (Optional[float]): Additional temperature change (DTADD).
            rate_mult (Optional[float]): Rate multiplier (RATEMULT).
            polymer (Optional[str]): Polymer identifier (POLYMER).
            station (Optional[str]): Station identifier (STATION).
            drill_queue (Optional[str]): Drill queue identifier (ASSCDR).
            drill_order_benefit (Optional[float]): Benefit of the drill order (BENEFIT).
            is_activated (bool): Whether the Well Connection has been activated using ACTIVATED / DEACTIVATE
            Defaults to True.
            bhdepth (float): Bottomhole depth (BHDEPTH)
            datum_depth (float): Depth relative to a datum (DATUM)
            x_pos (float): X-coordinate position (X)
            y_pos (float): Y-coordinate position (Y)
            length (float): Length of the well connection (LENGTH)
            temperature (float): Temperature (TEMP)
            diameter (float): Diameter of the well connection (DIAM)
            roughness (float): Pipe roughness (ROUGHNESS)
            inner_diameter (float): Inner diameter of the well connection (INNERDIAM)
            productivity_index_phase (str): Productivity index phase identifier (PIPHASE)
            hyd_method (str): Hydraulic method (METHOD)
            group (Optional[str]): The group that the Well Connection belongs to.
            drainage_radius (Optional[float]): The drainage radius.
            i (Optional[int]): The location of the Well Connection in the i direction.
            j (Optional[int]): The location of the Well Connection in the j direction.
            productivity_index (Optional[float]): Productivity index (PI).
        """

        self.bh_node_name = bh_node_name
        self.wh_node_name = wh_node_name
        self.stream = stream
        self.number = number
        self.scale = scale
        self.gradient_calc = gradient_cal
        self.bottomhole_measured_depth = bottomhole_measured_depth
        self.add_tubing = add_tubing
        self.tracer = tracer
        self.con_type = con_type
        self.pvt_method = pvt_method
        self.water_method = water_method
        self.bat_method = bat_method
        self.elevation_profile = elevation_profile
        self.temperature_profile = temperature_profile
        self.inj_mobility = inj_mobility
        self.crossshut = crossshut
        self.crossflow = crossflow
        self.on_time = on_time
        self.heat_transfer_coeff = heat_transfer_coeff
        self.well_index_mult = well_index_mult
        self.vip_productivity_index = vip_productivity_index
        self.productivity_index_phase = productivity_index_phase
        self.d_factor = d_factor
        self.non_darcy_flow_model = non_darcy_flow_model
        self.non_darcy_flow_method = non_darcy_flow_method
        self.gas_mobility = gas_mobility
        self.capillary_number_model = capillary_number_model
        self.dp_add = dp_add
        self.dt_add = dt_add
        self.rate_mult = rate_mult
        self.polymer = polymer
        self.station = station
        self.drill_queue = drill_queue
        self.drill_order_benefit = drill_order_benefit
        super().__init__(date_format=date_format, start_date=start_date, unit_system=unit_system, name=name, date=date,
                         properties_dict=properties_dict, bhdepth=bhdepth, datum_depth=datum_depth, x_pos=x_pos,
                         y_pos=y_pos, length=length, temperature=temperature, diameter=diameter, roughness=roughness,
                         inner_diameter=inner_diameter, productivity_index=productivity_index, hyd_method=hyd_method,
                         group=group, i=i, j=j, drainage_radius=drainage_radius)

        if self.name is not None:
            self.bh_node_name = self.name + '%bh'
            self.wh_node_name = self.name + '%wh'

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'NAME': ('name', str),
            'STREAM': ('stream', str),
            'NUMBER': ('number', int),
            'SCALE': ('scale', float),
            'BHDEPTH': ('bhdepth', float),
            'DATUM': ('datum_depth', float),
            'X': ('x_pos', float),
            'Y': ('y_pos', float),
            'DATGRAD': ('gradient_calc', str),
            'LENGTH': ('length', float),
            'BHMD': ('bottomhole_measured_depth', float),
            'ADDTUBING': ('add_tubing', int),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'TRACERS': ('tracer', str),
            'TYPE': ('con_type', str),
            'METHOD': ('hyd_method', str),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int),
            'IBAT': ('bat_method', int),
            'TEMP': ('temperature', float),
            'ELEVPR': ('elevation_profile', str),
            'TEMPPR': ('temperature_profile', str),
            'INJMOB': ('inj_mobility', str),
            'CROSS_SHUT': ('crossshut', str),
            'CROSSFLOW': ('crossflow', str),
            'ONTIME': ('on_time', float),
            'HTC': ('heat_transfer_coeff', float),
            'WIMULT': ('well_index_mult', float),
            'PI': ('productivity_index', float),
            'VIPPI': ('vip_productivity_index', float),
            'PIPHASE': ('productivity_index_phase', str),
            'D': ('d_factor', float),
            'ND': ('non_darcy_flow_model', str),
            'DPERF': ('non_darcy_flow_method', str),
            'GASMOB': ('gas_mobility', str),
            'CN': ('capillary_number_model', str),
            'DPADD': ('dp_add', float),
            'DTADD': ('dt_add', float),
            'RATEMULT': ('rate_mult', float),
            'POLYMER': ('polymer', str),
            'STATION': ('station', str),
            'ASSCDR': ('drill_queue', str),
            'BENEFIT': ('drill_order_benefit', float),
            }
        return nexus_mapping

    def __repr__(self) -> str:
        return super().__repr__()
